from . import _functional as F
from ..data import Context, SharedBase
from ..utils import get_member
import typing
import functools


def get_instance(instance, is_method, args):
    """Get the instance

    Args:
        instance: _description_
        is_method (bool): whether the function is a method
        args: The args for the insta

    Returns:
        the instance and args
    """
    if is_method and instance is None:
        return args[0], args[1:]
    
    return instance, args
    

class CompositeFunc:
    """CompositeFunc is used for decorating sequences and selectors
    """

    def __init__(
        self, f, base_f, ctx: 
        Context=None, is_method: bool=False, instance=None
    ):
        """Create a composite function

        Args:
            f: The function to execute
            base_f: The base composite function to use (i.e. selector or sequence)
            ctx (Context, optional): The context for the function. Defaults to None.
            is_method (bool): Whether it is a method or not
            instance (optional): The instance. Defaults to None.
        """
        self.f = f
        self.base_f = base_f
        self.instance = instance
        self.is_method = is_method
        self._ctx = ctx

    def task(self, *args, _ctx: Context=None, **kwargs):
        """Get the task for the function

        Args:
            _ctx (Context, optional): An override for the context. Defaults to None.

        Returns:
            The task 
        """
        # This method will handle "task" and correctly bind to the instance

        instance, args = get_instance(self.instance, self.is_method, args)
        
        ctx = _get(instance, self._ctx, _ctx)  
        if instance is None:
            return self.base_f(self.f, ctx, *args, **kwargs)
        
        return self.base_f(self.f, ctx, instance, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        """Execute the function

        Returns: The output of the function
        """
        # This handles the original method call
        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is not None:
            return self.f(instance, *args, **kwargs)
        return self.f(*args, **kwargs)

    def __get__(self, instance, owner):

        if self.f.__name__ in instance.__dict__:
            return instance.__dict__[self.f.__name__]
        
        task = CompositeFunc(self.f, self.base_f, self._ctx, True, instance)
        instance.__dict__[self.f.__name__] = task
        return task


class ParallelFunc:
    """TaskF is used for function decorators to 
    """

    def __init__(
        self, f, succeeds_on=-1, fails_on=1, success_priority=True,
        is_method: bool=False,
        instance=None
    ):
        """A Parallel Function decorator executes a parallel task

        Args:
            f : The function to execute
            ctx (Context, optional): The context to use. Defaults to None.
            succeeds_on (int): the number of successes before regarded as success
            fails_on (int): the number of failures before it is regarded as failing
            success_priority (int): whether it success is prioritized over failure
            is_method (bool): 
            instance (optional): The instance. Defaults to None.
        """
        self.f = f
        self.instance = instance
        self.succeeds_on = succeeds_on
        self.fails_on = fails_on
        self.is_method = is_method
        self.success_priority = success_priority

    def task(self, *args, **kwargs):
        """Get the task from the function

        Returns:
            The task to exeucte
        """
        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is None:
            return F.parallelf(
                self.f, *args, succeeds_on=self.succeeds_on, 
                fails_on=self.fails_on, success_priority=self.success_priority, **kwargs
            )
        
        return F.parallelf(
            self.f, instance, *args, 
            succeeds_on=self.succeeds_on, fails_on=self.fails_on, success_priority=self.success_priority, **kwargs
        )

    def __call__(self, *args, **kwargs):
        """Execute the function

        Returns: The output of the function
        """
        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is not None:
            return self.f(instance, *args, **kwargs)
        return self.f(*args, **kwargs)

    def __get__(self, instance, owner):
        """Add the task to the instance if not already there
        """

        if self.f.__name__ in instance.__dict__:
            return instance.__dict__[self.f.__name__]
        
        task = ParallelFunc(
            self.f, self.succeeds_on, self.fails_on, 
            self.success_priority, True, instance
        )
        instance.__dict__[self.f.__name__] = task
        return task


class CondFunc:
    """TaskF is used for function decorators to 
    """

    def __init__(
        self, f, is_method: bool=False, instance=None
    ):
        """Create a conditional function

        Args:
            f: The conditional function
            instance (optional): _description_. Defaults to None.
        """
        self.f = f
        self.instance = instance
        self.is_method = is_method

    def task(self, *args, **kwargs):
        """Get the task from the function

        Returns:
            The task to exeucte
        """
        # This method will handle "task" and correctly bind to the instance

        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is None:
            return F.condf(self.f, *args, **kwargs)
        return F.condf(self.f, instance, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        """Execute the function

        Returns: The output of the function
        """
        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is not None:
            return self.f(instance, *args, **kwargs)
        return self.f(*args, **kwargs)

    def __get__(self, instance, owner):
        """Add the task to the instance if not already there
        """
        if self.f.__name__ in instance.__dict__:
            return instance.__dict__[self.f.__name__]
        task = CondFunc(self.f, True, instance)
        instance.__dict__[self.f.__name__] = task
        return task


def _get_str(obj, key):

    return get_member(obj, key) if isinstance(key, str) else key
            

class TaskFunc:
    """TaskF is used for function decorators to 
    """

    def __init__(
        self, f, is_method: bool=False, instance=None, 
        out: typing.Union[SharedBase, str]=None, 
        to_status: F.TOSTATUS=None
    ):
        """A general decorator for a Task Function

        Args:
            f: The task function
            is_method (bool, optional): Whether it is a method or not. Defaults to False.
            instance (_type_, optional): The instance for the task. Defaults to None.
            out (typing.Union[SharedBase, str], optional): The processor for the result. Defaults to None.
            to_status (F.TOSTATUS, optional): How to convert to a status. Defaults to None.
        """
        self.f = f
        self.instance = instance
        self.is_method = is_method
        self._to_status = to_status
        self._out = out

    def task(self, *args, **kwargs):
        """Get the "task" for the function

        Returns:
            The task
        """

        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is None:
            out = self._out
            to_status = self._to_status
            return F.taskf(self.f, *args, out=out, to_status=to_status, **kwargs)
        else:
            to_status = _get_str(instance, self._to_status)
            out = _get_str(instance, self._out)
            return F.taskf(
                self.f, instance, out=out, to_status=to_status, 
                *args, **kwargs
            )

    def __call__(self, *args, **kwargs):
        instance, args = get_instance(self.instance, self.is_method, args)

        if instance is not None:
            return self.f(instance, *args, **kwargs)
        return self.f(*args, **kwargs)

    def __get__(self, instance, owner):
        """Add the task to the instance if not already there
        """

        if self.f.__name__ in instance.__dict__:
            return instance.__dict__[self.f.__name__]
        
        task = TaskFunc(self.f, True, instance, self._out, self._to_status)
        instance.__dict__[self.f.__name__] = task
        return task


def _get(self, override, base):
    """Get the an override value if defined

    Args:
        override: The override value
        base: The base value

    Raises:
        ValueError: If neither value has been set

    Returns:
        Any
    """
    
    base = base or override
    if base is None:
        raise ValueError('Value has not been defined')
    elif isinstance(base, str):
        return get_member(self, base)
    return base


def sequencefunc(ctx: Context=None, is_method: bool=False):
    """Decorate a sequence function that yields tasks

    Args:
        ctx (Context, optional): The context. Defaults to None.
        is_method (bool, optional): Whether it is a method. Defaults to False.
    Returns: The task
    """
    def _(f):

        return CompositeFunc(f, F.sequencef, ctx, is_method)
    return _


def sequencemethod(ctx: Context):
    """Decorate a sequence method that yields tasks

    Args:
        ctx (Context): 

    Returns: The task
    """
    return sequencefunc(ctx, True)


def selectorfunc(ctx: Context=None):
    """Decorate a selector function that yields tasks

    Args:
        ctx (Context, optional): The context for the task. Defaults to None.

    Returns: the task
    """
    def _(f):
        return CompositeFunc(f, F.selectorf, ctx)
    return _


def selectormethod(ctx: Context):
    """Decorate a selector method that yields tasks

    Args:
        ctx (Context): The context for the task

    Returns: the task
    """
    return selectorfunc(ctx, True)


fallbackfunc = selectorfunc
fallbackmethod = selectormethod


def parallelfunc(
    succeeds_on: int=-1, fails_on: int=1, 
    success_priority: bool=True, is_method: bool=False
):
    """Decorate a parallel function that yields tasks

    Args:
        succeeds_on (int, optional): Number required to succeed. Defaults to -1.
        fails_on (int, optional): Number required to fail. Defaults to 1.
        success_priority (bool, optional): Whether success prioritized over failure if equal. Defaults to True.
        is_method (bool, optional): Whether it is a method. Defaults to False.
    """

    def _(f):
        return ParallelFunc(f, succeeds_on, fails_on, success_priority, is_method)
    return _


def parallelmethod(
    succeeds_on: int=-1, fails_on: int=1, 
    success_priority: bool=True
):
    """Decorate a parallel method that yields tasks

    Args:
        succeeds_on (int, optional): Number required to succeed. Defaults to -1.
        fails_on (int, optional): Number required to fail. Defaults to 1.
        success_priority (bool, optional): Whether success prioritized over failure if equal. Defaults to True.
    """
    return parallelfunc(
        succeeds_on, fails_on, success_priority, True
    )


def condfunc(is_method: bool=False):
    """Decorate a conditional function that returns True or False

    Args:
        is_method (bool, optional): Whether it is a method. Defaults to False.
    """
    def _(f):
        return CondFunc(f, is_method)
    return _


def condmethod():
    """Decorate a conditional method
    """
    return condfunc(True)


def taskfunc(out: SharedBase=None, to_status: F.TOSTATUS=None, is_method: bool=False):
    """Decorate a general task function

    Args:
        out (SharedBase, optional): The processor for the result. Defaults to None.
        to_status (F.TOSTATUS, optional): The converter to change the result to a status. Defaults to None.
        is_method (bool, optional): Whether it is a method or not. Defaults to False.
    """
    def _(f):
        t = TaskFunc(f, is_method, F.taskf, out, to_status)
        t.__call__ = functools.wraps(f, t.__call__)
        return t
    return _


def taskmethod(out: SharedBase=None, to_status: F.TOSTATUS=None):
    """Decorate a general task method

    Args:
        out (SharedBase, optional): The processor for the result. Defaults to None.
        to_status (F.TOSTATUS, optional): The converter to change the result to a status. Defaults to None.
    """
    return taskfunc(
        out, to_status, True
    )



from ._utils import (
    get_str_variables, escape_curly_braces,
    unescape_curly_braces, is_primitive,
    generic_class, str_formatter,
    is_nested_model, is_undefined,
    UNDEFINED, WAITING,
    primitives, get_member
)
from ._model import (
    Renderable, model_template,
    struct_template,
    model_to_text, model_from_text,
    StructLoadException, Templatable,
    TemplateField, doc
)



# 1st party
import typing
from functools import wraps, update_wrapper
import inspect
from itertools import chain

import pydantic

# local
from ._core import (
    render, 
    Cue, render, Param, 
    Instruct, Reader, NullRead,
)
from .._core._ai import (
    Dialog, TextMessage, AIModel, AIResponse
)
from ._read import (
    PydanticRead, PrimRead
)
from ..utils._utils import (
    str_formatter, primitives, get_member
)
from .._core._process import Module


X = typing.Union[str, Cue]


def validate_out(cues: typing.List[X]) -> typing.Optional[Reader]:
    """Validate an Out based on several instructions

    Args:
        instructions (typing.List[X]): The instructions 

    Returns:
        Out: The resulting "Out" to use from the instructions
    """
    if isinstance(cues, dict):
        cues = cues.values()

    out = None
    for cue in cues:
        if not isinstance(cue, Cue):
            continue
        if out is None and cue.out is not None:
            out = cue.out
        elif cue.out is not None:
            raise RuntimeError(f'Out cannot be duplicated')
    return out


class InstructCall(Module, Instruct):
    """InstructCall is used within an instructf or
    signaturef to allow one to loop over "incoming"
    instructions
    """

    def __init__(self, i: Instruct, *args, **kwargs):
        """Create the "Call" passing in the cue
        and its inputs

        Args:
            i (Instruct): The cue to wrap
        """
        self._instruct = i
        self.args = args
        self.kwargs = kwargs

    def __iter__(self) -> typing.Iterator['InstructCall']:
        """Loop over the "sub" InstructCalls

        Yields:
            typing.Iterator['InstructCall']: The InstructCalls used
        """
        for arg in chain(self.args, self.kwargs.values()):
            if isinstance(arg, InstructCall):
                for arg_i in arg:
                    yield arg_i
                yield arg
        yield self

    def i(self) -> Cue:
        """Get the cue 

        Returns:
            Cue: The cue
        """
        return self._instruct.i()
        
    def forward(self) -> typing.Any:
        """Execute the cue

        Returns:
            typing.Any: Execute the Instruct
        """
        
        args = [
            arg() if isinstance(arg, InstructCall) 
            else arg for arg in self.args
        ]
        kwargs = {
            k: v() if isinstance(v, Instruct) else v
            for k, v in self.kwargs.items()
        }

        return self._instruct(*args, **kwargs)


class SignatureFunc(Module, Instruct):
    """SignatureFunc is a method where you define the cue in
    the function signature
    """
    def __init__(
        self, f: typing.Callable, engine: typing.Union[AIModel, str, typing.Callable[[], AIModel]], 
        dialog_factory: typing.Optional[typing.Callable[[], Dialog]]=None,
        doc: typing.Optional[str]=None,
        reader: typing.Optional[Reader]=None,
        train: bool=False, 
        ai_kwargs: typing.Dict=None,
        is_method: bool=False,
        instance=None
    ):
        """Wrap the signature method with a particular engine and
        dialog factory

        Args:
            f (typing.Callable): The function to wrap
            engine (AIModel): The engine to use for getting the response
            dialog_factory (typing.Optional[typing.Callable[[], Dialog]], optional): The dialog to use. Defaults to None.
            train (bool, optional): Whether to train the cues or not. Defaults to False.
            instance (optional): The instance. Defaults to None.
        """
        self.f = f
        self.name = f.__name__
        self.engine = engine
        self._train = train
        self._is_method = is_method
        self._doc = doc
        docstring = inspect.getdoc(f) if doc is None else doc
        self.signature = str(inspect.signature(f))
        self.parameters = inspect.signature(f).parameters
        self.return_annotation = inspect.signature(f).return_annotation

        if not isinstance(docstring, typing.Callable):
            docstring = Cue(text=docstring)
            self._docstring = Param(
                name=self.name,
                cue=docstring,
                training=train
            )
        elif train:
            raise ValueError('Cannot set to train if the docstring is a callable')
        else:
            self._docstring = docstring

        self.out_cls = (
            self.return_annotation if self.return_annotation is not None 
            else AIResponse 
        )

        if reader is None:
            if self.out_cls in primitives:
                reader = PrimRead(name=self.name, out_cls=self.out_cls)
            elif issubclass(self.out_cls, pydantic.BaseModel):
                reader = PydanticRead(name=self.name, out_cls=self.out_cls)
            else:
                reader = NullRead(name=self.name)
        
        self.reader = reader

        update_wrapper(self, f) 
        self.dialog_factory = dialog_factory or Dialog
        self.ai_kwargs = ai_kwargs
        self._instance = instance

    def spawn(
        self, 
        engine: AIModel=None, 
        dialog_factory: typing.Optional[typing.Callable[[], Dialog]]=None,
        train: bool=False
    ) -> 'SignatureFunc':
        """Spawn a new SignatureMethod. Especially use to create a trainable one

        Args:
            engine (AIModel, optional): Spawn a new . Defaults to None.
            dialog_factory (typing.Optional[typing.Callable[[], Dialog]], optional): _description_. Defaults to None.
            train (bool, optional): _description_. Defaults to False.

        Returns:
            SignatureMethod: 
        """
        return SignatureFunc(
            f=self.f,
            engine=engine or self.engine,
            dialog_factory=dialog_factory or self.dialog_factory,
            reader=self.reader,
            doc=self._doc,
            train=train,
            instance=self._instance,
            ai_kwargs=self.ai_kwargs
        )

    def i(self, *args, **kwargs) -> Cue:
        """Get the cue

        Returns:
            Cue: Get the cue
        """

        if self._instance is not None:
            instance = self._instance
        elif self._is_method:
            instance = args[0]
            args = args[1:]
        else:
            instance = None
        
        if isinstance(self._docstring, Param):
            filled_docstring = self._docstring.render()
        else:
            filled_docstring = self._docstring()

        filled = set()

        param_values = list(self.parameters.values())

        if isinstance(self.reader, str) and instance is not None:
            reader = get_member(instance, self.reader)
        else:
            reader = self.reader

        if instance is not None:
            values = self.f(instance, *args, **kwargs)
            param_values = param_values[1:]
        else:
            values = self.f(*args, **kwargs)
        values = values if values is not None else {}
        values = {k: v() if isinstance(k, InstructCall) else v for k, v in values.items()}

        # what if one of the parameters is an cue?
        # values.update(dict(zip(args, [v for v in param_values])))
        
        if '{TEMPLATE}' in filled_docstring:
            values['TEMPLATE'] = reader.template()

        for value, param in zip(args, param_values):
            values[param.name] = value
            filled.add(param.name)
        
        for k, value in kwargs.items():
            param = self.parameters[k]
            values[param.name] = value
            filled.add(param.name)

        for param in param_values:
            if param.name in filled:
                continue
            if param.default == inspect.Parameter.empty:
                raise RuntimeError('Param has not been defined and no value')
            
            values[param.name] = param.default

        # TODO: Determine how to handle this
        out = validate_out(values)
        values = {key: render(v) for key, v in values.items()}
        
        print(filled_docstring, values.keys())
        filled_docstring = str_formatter(
            filled_docstring, required=False, **values
        )

        return Cue(
            text=filled_docstring,
            out=reader, 
            # out=StructFormatter(name=self.name, out_cls=self.out_cls)
        )

    def forward(
        self, *args,
        _engine: AIModel=None, **kwargs
    ) -> typing.Any:
        """Execute the cue and get the output 

        Args:
            _engine (AIModel, optional): The engine to override with. Defaults to None.

        Returns:
            typing.Any: The result of processing the cue
        """
        engine = _engine or self.engine

        if self._instance is not None:
            instance = self._instance
        elif self._is_method:
            instance = args[0]

        if isinstance(engine, str) and instance is not None:
            engine = get_member(instance, engine)
        elif not isinstance(engine, AIModel) and isinstance(engine, typing.Callable):
            engine = engine()

        cue = self.i(*args,  **kwargs)
        result = engine(TextMessage('system', cue), **self.ai_kwargs)
        if self.out_cls is AIResponse:
            return result
        return result.val

    def stream_forward(
        self, *args,
        _engine: AIModel=None, **kwargs
    ) -> typing.Iterator[typing.Tuple[typing.Any, typing.Any]]:
        """Execute the cue and get the output 

        Args:
            _engine (AIModel, optional): The engine to override with. Defaults to None.

        Returns:
            typing.Any: The result of processing the cue
        """
        engine = _engine or self.engine

        if self._instance is not None:
            instance = self._instance
        elif self._is_method:
            instance = args[0]

        if isinstance(engine, str) and instance is not None:
            engine = get_member(instance, engine)
        elif not isinstance(engine, AIModel) and isinstance(engine, typing.Callable[[], AIModel]):
            engine = engine()

        cue = self.i(*args,  **kwargs)
        for cur, dx in engine.stream_forward(TextMessage('system', cue), **self.ai_kwargs):

            if self.out_cls is AIResponse:
                yield cur, dx
            else:
                yield cur.val, dx.val

    async def async_forward(
        self, *args, 
        _engine: AIModel=None, 
        **kwargs
    ) -> typing.Any:
        """Execute the cue and get the output

        Args:
            _engine (AIModel, optional): The engine to override with. Defaults to None.

        Returns:
            typing.Any: The result of processing the cue
        """
        return self.forward(
            *args, 
            _engine=_engine,
            **kwargs
        )

    def __iter__(self, *args, **kwargs) -> typing.Iterator[InstructCall]:
        """Loop over all child InstructCalls of this "Instruct"

        Yields:
            InstructCall
        """
        res = self(*args, **kwargs)
        for k, v in res.items():
            if isinstance(v, InstructCall):
                for v_i in v:
                    yield v_i
        yield InstructCall(self, *args, **kwargs)

    def __get__(self, instance, owner):
        """Set the instance on the SignatureMethod

        Args:
            instance (): The instance to use
            owner (): 

        Returns:
            SignatureMethod
        """
        if self.f.__name__ not in instance.__dict__:
            instance.__dict__[self.f.__name__] = SignatureFunc(
                self.f, self.engine, self.dialog_factory,
                self._doc, self.reader, self._train,
                self.ai_kwargs, True, instance
            )
        return instance.__dict__[self.f.__name__]


class InstructFunc(Module, Instruct):
    """InstructMethod is a method where you define the cue by
    doing operations on that instructions
    """

    def __init__(
        self, f: typing.Callable, engine: typing.Union[AIModel, str, typing.Callable[[], AIModel]], 
        dialog_factory: typing.Optional[typing.Callable[[], Dialog]]=None,
        ai_kwargs=None,
        is_method: bool=False,
        instance=None
    ):
        """Create an InstructMethod that decorates a function that returns 
        a cue

        Args:
            f (typing.Callable): The function to decorate
            train (bool, optional): Whether to train . Defaults to True.
            instance (, optional): The instance to use. Defaults to None.

        """
        self.f = f
        self.engine = engine
        update_wrapper(self, f) 
        self._instance = instance
        self._stored = None
        self._is_method = is_method
        self.dialog_factory = dialog_factory or Dialog
        self.return_annotation = inspect.signature(f).return_annotation

        # make it so it can automatically set this up
        # rather than using the "Null version"
        # if reader is None:
        #     if self.out_cls in primitives:
        #         reader = PrimRead(name=self.name, out_cls=self.out_cls)
        #     elif issubclass(self.out_cls, Struct):
        #         reader = StructRead(name=self.name, out_cls=self.out_cls)
        #     else:
        #         reader = NullRead(name=self.name)
        
        # self.reader = reader or NullRead()
        self.ai_kwargs = ai_kwargs or {}

    def i(self, *args, **kwargs) -> Cue:
        """Get the cue based on the arguments

        Returns:
            Cue: Get the cue
        """

        if self._instance is not None:
            instance = self._instance
        elif self._is_method:
            instance = args[0]
            args = args[1:]
        else:
            instance = None

        if instance is not None:
            result = self.f(instance, *args, **kwargs)
        else:
            result = self.f(*args, **kwargs)
    
        if isinstance(result, InstructCall):
            result = result()
        return result

    def forward(self, *args, _engine: AIModel=None, **kwargs) -> typing.Any:        
        """Execute the instruct method and then process the output

        Args:
            _engine (AIModel, optional): Engine to override with. Defaults to None.

        Returns:
            typing.Any: The resulting
        """
        if self._instance is not None:
            instance = self._instance
        elif self._is_method:
            instance = args[0]

        engine = _engine or self.engine
        if isinstance(engine, str) and instance is not None:
            engine = get_member(instance, engine)
        elif not isinstance(engine, AIModel) and isinstance(engine, typing.Callable):
            engine = engine()

        cue = self.i(*args, **kwargs)
        result = engine(TextMessage('system', cue), **self.ai_kwargs)
        if self.return_annotation is AIResponse:
            return result
        return result.val

    def stream_forward(
        self, *args,
        _engine: AIModel=None, **kwargs
    ) -> typing.Iterator[typing.Tuple[typing.Any, typing.Any]]:
        """Execute the cue and get the output 

        Args:
            _engine (AIModel, optional): The engine to override with. Defaults to None.

        Returns:
            typing.Any: The result of processing the cue
        """
        engine = _engine or self.engine

        if self._instance is not None:
            instance = self._instance
        elif self._is_method:
            instance = args[0]

        if isinstance(engine, str) and self._instance is not None:
            engine = get_member(instance, engine)
        
        elif not isinstance(engine, AIModel) and isinstance(engine, typing.Callable[[], AIModel]):
            engine = engine()

        cue = self.i(*args,  **kwargs)
        for cur, dx in engine.stream_forward(TextMessage('system', cue), **self.ai_kwargs):

            if self.return_annotation is AIResponse:
                yield cur, dx
            else:
                yield cur.val, dx.val

    async def async_forward(self, *args, **kwargs) -> typing.Any:
        """Execute forward asynchronously

        Returns:
            typing.Any: The result of the forward method
        """
        # TODO: Update this to use the Async for the Engine
        return self.forward(*args, **kwargs)

    def __get__(self, instance, owner):
        """Get the SignatureMethod with the instance specified

        Args:
            instance (): The instance to use
            owner (): 

        Returns:
            SignatureMethod
        """
        if self.f.__name__ not in instance.__dict__:
            instance.__dict__[self.f.__name__] = InstructFunc(
                self.f, self.engine, self.dialog_factory,
                self.ai_kwargs, True, instance
            )
        return instance.__dict__[self.f.__name__]

    def __iter__(self, *args, **kwargs):
        """Loop over all child InstructCalls of this "Instruct"

        Yields:
            InstructCall
        """
        res = self(*args, **kwargs)
        if isinstance(res, InstructCall):
            for res_i in res:
                yield res_i


def instructfunc(
    engine: AIModel=None,
    is_method: bool=False,
    **ai_kwargs
):
    """Decorate a method with instructfunc

    Args:
        engine (AIModel, optional): The engine for the AI . Defaults to None.
        is_method (bool): Whether it is a method

    Returns:
        typing.Callable[[function], SignatureFunc]
    """
    def _(f):

        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)
        
        # TODO: Use wrapper
        return InstructFunc(
            f, engine, None, is_method=is_method,
            ai_kwargs=ai_kwargs
        )
    return _


def instructmethod(
    engine: AIModel=None,
    **ai_kwargs
):
    """Decorate a method with instructfunc

    Args:
        engine (AIModel, optional): The engine for the AI . Defaults to None.

    Returns:
        typing.Callable[[function], SignatureFunc]
    """
    return instructfunc(
        engine, True, **ai_kwargs
    )


def signaturefunc(
    engine: AIModel=None, 
    reader: Reader=None,
    doc: typing.Union[str, typing.Callable[[], str]]=None,
    is_method=False,
    **ai_kwargs
):
    """Decorate a method with SignatureFunc

    Args:
        engine (AIModel, optional): The engine for the AI . Defaults to None.
        reader (Reader, optional): The reader to use for the method. Defaults to None.
        doc (typing.Union[str, typing.Callable[[], str]], optional): A docstring to override with. Defaults to None.
        is_method (bool): Whether the function is a method. 

    Returns:
        typing.Callable[[function], SignatureFunc]
    """
    def _(f):

        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)
        
        return SignatureFunc(
            f, engine, None, is_method=is_method, doc=doc,
            reader=reader, ai_kwargs=ai_kwargs
        )

    return _


def signaturemethod(
    engine: AIModel=None, 
    reader: Reader=None,
    doc: typing.Union[str, typing.Callable[[], str]]=None,
    **ai_kwargs
):
    """Decorate a method with SignatureFunc

    Args:
        engine (AIModel, optional): The engine for the AI . Defaults to None.
        reader (Reader, optional): The reader to use for the method. Defaults to None.
        doc (typing.Union[str, typing.Callable[[], str]], optional): A docstring to override with. Defaults to None.

    Returns:
        typing.Callable[[function], SignatureFunc]
    """
    return signaturefunc(
        engine, reader, doc, True, **ai_kwargs
    )



# 1st party
from abc import abstractmethod
import typing
import time
import random

# 3rd party
from . import _functional
from ._core import Task, TaskStatus, State
from ..data._data import Context

from ..data import Buffer
from .._core import AIModel


class Root(Task):
    """The root task for a behavior tree
    """

    def __init__(self, root: 'Task'=None):
        """Create a tree to store the tasks

        Args:
            root (Task, optional): The root task. Defaults to None.
            name (str): The name of the tree
        """
        super().__init__()
        self._root = root

    def tick(self) -> TaskStatus:
        """Update the task

        Returns:
            SangoStatus: The status after tick
        """
        if self._root is None:
            return TaskStatus.SUCCESS

        return self._root.tick()

    @property
    def root(self) -> Task:
        """
        Returns:
            Task: The root task
        """
        return self._root
    
    @root.setter
    def root(self, root: 'Task'):
        """
        Args:
            root (Task): The root task
        """
        self._root = root

    def reset(self):
        """Reset the status of the task

        """
        super().reset()
        if self._root is not None:
            self._root.reset()


class Serial(Task):
    """A task consisting of other tasks executed one 
    after the other
    """

    def __init__(self, tasks: typing.Iterable[Task]):
        """Create a serial task

        Args:
            tasks (typing.Iterable[Task]): The tasks comprising the serial task
        """
        super().__init__()
        self._tasks = tasks
        self._context = Context()
        
    @property
    def tasks(self) -> typing.Iterable[Task]:
        """Get the tasks in the serial task

        Returns:
            typing.Iterable[Task]: The tasks comprising the serial task
        """
        return self._tasks

    def reset(self):
        """Reset the state
        """
        super().reset()
        self._context = Context()
        for task in self._tasks:
            task.reset()


class Sequence(Serial):
    """Create a sequence of tasks to execute
    """

    def __init__(self, tasks: typing.Iterable[Task]):
        """Create a sequence of tasks

        Args:
            tasks (typing.Iterable[Task]): The tasks making up the sequence
        """
        super().__init__(tasks)
        self.f = _functional.sequence(
            self._tasks, self._context
        )

    def tick(self) -> TaskStatus:
        """Update the task

        Returns:
            TaskStatus: The status
        """
        self._status = self.f()
        return self._status

    def reset(self):
        super().reset()
        self.f = _functional.sequence(
            self._tasks, self._context
        )


class Selector(Serial):
    """Create a set of tasks to select from
    """

    def __init__(self, tasks: typing.Iterable[Task]):
        """Create a selector of tasks

        Args:
            tasks (typing.Iterable[Task]): The tasks to select from
        """
        super().__init__(tasks)
        self.f = _functional.selector(
            self._tasks, self._context
        )

    def tick(self) -> TaskStatus:
        """Execute the task

        Returns:
            TaskStatus: The resulting task status
        """
        self._status = self.f()
        return self._status

    def reset(self):
        super().reset()
        self.f = _functional.selector(
            self._tasks, self._context
        )


Fallback = Selector


class Parallel(Task):
    """A composite task for running multiple tasks in parallel
    """

    def __init__(self, tasks: typing.Iterable[Task]=None, fails_on: int=None, succeeds_on: int=None, success_priority: bool=True):
        """The parallel

        Args:
            tasks (typing.Iterable[Task]): 
            runner (, optional): . Defaults to None.
            fails_on (int, optional): . Defaults to None.
            succeeds_on (int, optional): . Defaults to None.
            success_priority (bool, optional): . Defaults to True.
        """
        super().__init__()
        self._tasks = tasks if tasks is not None else []
        self._fails_on = fails_on if fails_on is not None else len(self._tasks)
        self._succeeds_on = succeeds_on if succeeds_on is not None else (len(self._tasks) + 1 - self._fails_on)

        self._success_priority = success_priority
        self._f = _functional.parallel(
            self._tasks, self._succeeds_on, self._fails_on,
            self._success_priority
        )
        self._ticked = set()

    def _update_f(self):
        self._f = _functional.parallel(
            self._tasks, self._succeeds_on, self._fails_on,
            self._success_priority
        )

    @property
    def tasks(self) -> typing.Iterable[Task]:
        return self._tasks

    def set_condition(self, fails_on: int, succeeds_on: int):
        """Set the number of falures or successes it takes to end

        Args:
            fails_on (int): The number of failures it takes to fail
            succeeds_on (int): The number of successes it takes to succeed

        Raises:
            ValueError: If teh number of successes or failures is invalid
        """
        self._fails_on = fails_on if fails_on is not None else len(self._tasks)
        self._succeeds_on = succeeds_on if succeeds_on is not None else (len(self._tasks) + 1 - self._fails_on)

    def validate(self):
        
        if (self._fails_on + self._succeeds_on - 1) > len(self._tasks):
            raise ValueError('')
        if self._fails_on <= 0 or self._succeeds_on <= 0:
            raise ValueError('')

    @property
    def fails_on(self) -> int:
        return self._fails_on

    @fails_on.setter
    def fails_on(self, fails_on) -> int:
        self._fails_on = fails_on
        self._update_f()
        return self._fails_on

    @property
    def succeeds_on(self) -> int:
        """Get the number required for success

        Returns:
            int: The number of successes to succeed
        """
        return self._succeeds_on        
    
    @succeeds_on.setter
    def succeeds_on(self, succeeds_on) -> int:
        """Set the number required for success

        Args:
            succeeds_on: the number required for success

        Returns:
            int: The number of successes to succeed
        """
        self._succeeds_on = succeeds_on
        self._update_f()
        return self._succeeds_on
    
    def tick(self) -> TaskStatus:
        """Tick the task

        Returns:
            TaskStatus: The status
        """
        self._status = self._f()
        return self._status

    def reset(self):
        """Reset the task and subtasks
        """

        super().reset()
        for task in self._ticked:
            task.reset()


class Action(Task):
    """A standard task that executes 
    """

    @abstractmethod
    def act(self) -> TaskStatus:
        """Commit an action

        Raises:
            NotImplementedError: 

        Returns:
            TaskStatus: The status of after executing
        """
        raise NotImplementedError

    def tick(self) -> TaskStatus:
        """Execute the action

        Returns:
            TaskStatus: The resulting status
        """
        if self._status.is_done:
            return self._status
        self._status = self.act()
        return self._status


class Condition(Task):
    """Check whether a condition is satisfied before
    running a task.
    """

    @abstractmethod
    def condition(self) -> bool:
        """Execute between

        Returns:
            bool: The result of the condition
        """
        pass

    def tick(self) -> TaskStatus:
        """Check the condition

        Returns:
            SangoStatus: Whether the condition failed or succeeded
        """
        self._status = TaskStatus.SUCCESS if self.condition() else TaskStatus.FAILURE
        return self._status


class Decorator(Task):

    # name should retrieve the name of the decorated
    def __init__(self, task: Task) -> None:
        """Decorate a task

        Args:
            task (Task): The task to decorate
        """
        super().__init__()
        self._task = task

    @abstractmethod
    def decorate(self, status: TaskStatus) -> bool:
        pass

    @property
    def task(self) -> Task:
        """Get the task to execute

        Returns:
            Task: The decorated task
        """
        return self._task

    def tick(self) -> TaskStatus:
        """Decorate the tick for the decorated task

        Args:
            terminal (Terminal): The terminal for the task

        Returns:
            SangoStatus: The decorated status
        """

        if self._status.is_done:
            return self._status

        self._status = self.decorate(
            self.task.tick()
        )
        return self._status

    def reset(self):
        """Reset the task and subtask
        """
        super().reset()
        self._task.reset()


class Until(Decorator):
    """Loop until a condition is met
    """
    def __init__(
        self, task: Task,
        target_status: TaskStatus= TaskStatus.SUCCESS
    ) -> None:
        """A decorator that will execute until a status is reached

        Args:
            task (Task): The sub task to run
            target_status (TaskStatus, optional): The status to break on. Defaults to TaskStatus.SUCCESS.
        """
        super().__init__(task)
        self.target_status = target_status

    def decorate(self, status: TaskStatus) -> TaskStatus:
        """Continue running unless the result is a success

        Args:
            status (SangoStatus): The status of the decorated task

        Returns:
            SangoStatus: The decorated status
        """
        if status == self.target_status:
            return status
        if status.is_done:
            self._task.reset()
        return TaskStatus.RUNNING


class Unless(Decorator):
    """Loop while a condition is met
    """

    def __init__(self, task: Task, target_status: TaskStatus= TaskStatus.FAILURE) -> None:
        """A decorator that will execute until a status is reached

        Args:
            task (Task): The sub task to run
            target_status (TaskStatus, optional): The status to break on. Defaults to TaskStatus.SUCCESS.
        """
        super().__init__(task)
        self.target_status = target_status

    def decorate(self, status: TaskStatus) -> TaskStatus:
        """Continue running unless the result is a failure

        Args:
            status (SangoStatus): The status of the decorated task

        Returns:
            SangoStatus: The decorated status
        """
        if status == self.target_status:
            return status
        if status.is_done:
            self._task.reset()
        return TaskStatus.RUNNING


class Not(Decorator):
    """Invert the result
    """

    def decorate(self, status: TaskStatus) -> TaskStatus:
        """Return Success if status is a Failure or Failure if it is a SUCCESS

        Args:
            status (SangoStatus): The status of the decorated task

        Returns:
            SangoStatus: The decorated status
        """
        return status.invert()


def run_task(task: Task, interval: typing.Optional[float]=1./60) -> typing.Iterator[TaskStatus]:
    """Run a task until completion

    Args:
        task (Task): The task to execute
        interval (float, optional): The interval to execute on. Defaults to 1./60.

    Yields:
        Iterator[typing.Iterator[TaskStatus]]: The status
    """

    status = None
    while status == TaskStatus.RUNNING or status == TaskStatus.READY:
        status = task.tick()
        if interval is not None:
            time.sleep(interval)
        yield status


class StateMachine(Task):
    """StateMachine is a task composed of multiple tasks in
    a directed graph
    """

    def __init__(self, init_state: 'State'):
        """Create the status

        Args:
            init_state (State): The starting state for the machine
        """
        self._init_state = init_state
        self._cur_state = init_state
        self._status = TaskStatus.READY

    def tick(self) -> TaskStatus:
        """Update the state machine
        """
        if self._status.is_done:
            return self._status
        
        self._cur_state = self._cur_state.update()
        if self._cur_state == TaskStatus.FAILURE or self._cur_state == TaskStatus.SUCCESS:
            self._status = self._cur_state
        else:
            self._status = TaskStatus.RUNNING
        return self._status

    @property
    def status(self) -> TaskStatus:
        """Get the status of the state machine

        Returns:
            TaskStatus: Get the current status
        """
        return self._status

    def reset(self):
        """Reset the state machine
        """
        super().reset()
        self._cur_state = self._init_state


class FixedTimer(Action):
    """A timer that will "succeed" at a fixed interval
    """

    def __init__(self, seconds: float):
        """Create the timer specifying the number of seconds for it

        Args:
            seconds (float): The number of seconds for the timer
        """
        super().__init__()
        self.seconds = seconds
        self._start = None

    def act(self) -> TaskStatus:
        """Execute the timer

        Returns:
            TaskStatus: The TaskStatus after running
        """
        cur = time.time()
        if self._start is None:
            self._start = cur
        elapsed = cur - self._start
        if elapsed >= self.seconds:
            return TaskStatus.SUCCESS
        return TaskStatus.RUNNING

    def reset(self):
        """Reset the timer
        """
        super().reset()
        self._start = None


class RandomTimer(Action):
    """A timer that will randomly choose a time between two values
    """
    def __init__(self, seconds_lower: float, seconds_upper: float):
        """Create the timer specifying the range

        Args:
            seconds_lower (float): The lower bound for the timer
            seconds_upper (float): The upper bound for the timer
        """
        super().__init__()
        self.seconds_lower = seconds_lower
        self.seconds_upper = seconds_upper
        self._start = None
        self._target = None

    def act(self) -> TaskStatus:
        """Execute the Timer

        Returns:
            TaskStatus: The status of the task
        """
        cur = time.time()
        if self._start is None:
            self._start = cur
            r = random.random() 
            self._target = r * self.seconds_lower + r * self.seconds_upper
        elapsed = cur - self._start
        if elapsed >= self._target:
            return TaskStatus.SUCCESS
        return TaskStatus.RUNNING

    def reset(self):
        """Reset the task
        """
        super().reset()
        self._start = None
        self._target = None



from ._core import (
    Task, TaskMessage, TaskStatus,
    Router, ROUTE, TOSTATUS, ToStatus, State,
    from_bool
)
from ._functional import (
    parallel, unless, until, sequence,
    action, not_, tick, condf, selector,
    unlessf, untilf, notf, parallelf,
    selectorf, sequencef, fallbackf,
    fallback, taskf, PARALLEL, threaded,
    stream_model, exec_model,
    exec_func
)
from ._tasks import (
    Serial, Sequence, 
    Selector, 
    Parallel, Action,
    Condition, Root,
    Not, 
    Unless, Until,
    run_task,
    Fallback,
    StateMachine,
    FixedTimer, RandomTimer
)
from ._build import (
    build_composite, build_sequence,
    build_parallel, build_sango, build_select, build_fallback,
    build_decorate, build_not, build_unless, build_until
)
from ._agent import (
    Agent, AgentStatus,
)
from ._decorator_func import (
    parallelfunc, selectorfunc, sequencefunc,
    taskfunc, condfunc
)



# 1st party
import typing
import json
import inspect

# 3rd party
from io import StringIO
import pandas as pd
import pydantic

# local
from .._core._core import (
    Reader, 
    render,
)
from ..utils import (
    struct_template,
    model_to_text,
    escape_curly_braces, 
)
from ..data._structs import DataList


S = typing.TypeVar('S', bound=pydantic.BaseModel)


class StructListRead(Reader, typing.Generic[S]):
    """Convert the output into a list of structs
    """
    name: str
    _out_cls: typing.Type[S] = pydantic.PrivateAttr()

    def __init__(self, out_cls: S, **data):
        """Create a converter specifying the struct class to convert
        to

        Args:
            out_cls (S): The class to convert to
        """
        super().__init__(**data)
        self._out_cls = out_cls

    def dump_data(self, data: DataList[S]) -> typing.Any:
        structs = []
        for cur in data.data:
            # 
            # structs.append(self._out_cls.to_dict(cur))
            structs.append(
                cur.model_dump()
            )

        return structs

    def write_text(self, data: typing.List) -> str:
        return json.dumps(data)

    def read_text(self, message: str) -> DataList[S]:
        """Convert the message into a list of structs

        Args:
            message (str): The AI message to read

        Returns:
            StructList[S]: the list of structs
        """
        return json.loads(message)
    
    def load_data(self, data) -> DataList[S]:
        structs = []
        for cur in data['data']:
            structs.append(self._out_cls(**cur))

        return DataList(data=structs)

    # TODO: Shouldn't this be "render?"
    def to_text(self, data: DataList[S]) -> str:
        """Convert the data to a string

        Args:
            data (StructList[S]): The data to convert to text

        Returns:
            str: the data converted to a string
        """
        return model_to_text(data)

    def template(self) -> str:
        """Output a template for the struct list

        Returns:
            str: A template of the struct list
        """
        # TODO: This is currently not correct
        #   Needs to output as a list

        return struct_template(self._out_cls)


class CSVRead(Reader):
    """Convert text to a StructList
    """
    indexed: bool = True
    delim: str = ','
    cols: typing.Optional[typing.Union[typing.Type[pydantic.BaseModel], typing.List[typing.Tuple[str, str, str]]]] = None

    def read_text(self, message: str):

        io = StringIO(message)
        df = pd.read_csv(io, sep=self.delim)
        return df.to_dict(orient='records', index=True)

    def load_data(self, data) -> typing.Dict:
        """Convert the message to a StructList

        Args:
            message (str): The message to convert

        Returns:
            StructList[S]: The result
        """
        return data # StructList[S].load_records(data)

    def dump_data(self, data: typing.List) -> typing.Any:
        return data

    def write_text(self, data: typing.Any) -> str:
        io = StringIO()
        data = [
            d_i.dump()
            for d_i in self.data.structs
        ]
        df = pd.DataFrame(data)
        df.to_csv(io, index=True, sep=self.delim)
        
        # Reset the cursor position to the beginning of the StringIO object
        io.seek(0)
        return io.read()

    def template(self) -> str:
        """Output a template for the CSV

        Returns:
            str: The template for the CSV
        """
        # s_cls: typing.Type[Struct] = generic_class(S)
        # template = s_cls.template()
        if (
            isinstance(self.cols, pydantic.BaseModel) or 
            (inspect.isclass(self.cols) and issubclass(self.cols, pydantic.BaseModel))
        ):
            temp = struct_template(self.cols)
            cols = []
            for k, v in temp.items():                
                # if 'description' not in v:
                #     raise RuntimeError(f'Cannot create CSV template for {self.cols}')
                cols.append((k, v.description, v.type_))
        elif self.cols is None:
            cols = [['1', 'First Col', ''], ['2', 'Second Col', ''], ['...', '...', ''], ['N', 'Nth Col', '']]
        else:
            cols = self.cols

        result = []
        header = ['Index']
        first = ['1']
        last = ['N']
        mid = '...'
        for name, descr, type_ in cols:
            header.append(name)
            first.append(f'{descr} <{type_}>')
            last.append(f'{descr} <{type_}>')
        header = f'{self.delim}'.join(header)
        first = f'{self.delim}'.join(first)
        last = f'{self.delim}'.join(last)
        result = [header, first, mid, last]
        result = '\n'.join(result)
        return result


class DualRead(Reader):

    text: Reader
    data: Reader

    def read_text(self, message: str) -> typing.Dict:
        return self.text.read_text(message)
    
    def load_data(self, data) -> typing.Any:
        return self.data.load_data(data)

    def dump_data(self, data: typing.Any) -> typing.Any:
        return self.data.dump_data(data)

    def write_text(self, data: typing.Any) -> str:
        return self.text.write_text(data)

    def template(self) -> str:
        return self.text.template()


class KVRead(Reader):
    """Create a Reader of a list of key values
    """

    sep: str = '::'
    key_descr: typing.Optional[typing.Union[typing.Type[pydantic.BaseModel], typing.Dict]] = None
    
    def read_text(self, message: str) -> typing.Dict:
        """Read in the list of key values

        Args:
            message (str): The message to read

        Returns:
            typing.Dict: A dictionary of keys and values
        """
        lines = message.splitlines()
        result = {}
        for line in lines:
            try:
                key, value = line.split(self.sep)
                result[key] = value
            except ValueError:
                pass
        return result
    
    def load_data(self, data: typing.Dict) -> typing.Dict:
        """Load data does not do anything as the result
        is a dictionary.

        Args:
            data (typing.Dict): The data to load

        Returns:
            typing.Dict: The dictionary of data
        """
        return data

    def dump_data(self, data: typing.Dict) -> typing.Dict:
        """Convert the data to a dictionary

        Args:
            data (typing.Dict): The data to load

        Returns:
            typing.Dict: The dumped data
        """
        return data

    def write_text(self, data: typing.Dict) -> str:
        """Write data as text

        Args:
            data (typing.Dict): The data to write

        Returns:
            str: The keys and values as text
        """
        return '\n'.join(
            f'{k}{self.sep}{render(v)}' for k, v in data.items()
        )

    def template(self) -> str:
        """Get the template for the Keys and Values

        Returns:
            str: The template
        """
        if self.key_descr is None:
            key_descr = {
                '<Example>': '<The value for the key.>'
            }
        elif (
            inspect.isclass(self.key_descr) and 
            issubclass(self.key_descr, pydantic.BaseModel)
        ):
            temp = struct_template(self.key_descr)
            key_descr = {}
            for k, v in temp.items():
                description =  v.description or 'value'
                key_descr[k] = f'<{description}>'
        else:
            key_descr = self.key_descr
        return '\n'.join(
            f'{key}{self.sep}{value}' 
            for key, value in key_descr.items()
        )


class JSONRead(Reader):
    """Use to read from a JSON
    """

    key_descr: typing.Optional[typing.Dict] = None

    def read_text(self, message: str) -> typing.Dict:
        """Read in the JSON

        Args:
            text (str): The JSON to read in

        Returns:
            typing.Dict: The result - if it fails, will return an empty dict
        """
        try: 
            result = json.loads(message)
            return result
        except json.JSONDecodeError:
            return {}
    
    def load_data(self, data: typing.Dict) -> typing.Dict:
        """Load the data from a dictionary. Since JSONs are just a dict
        this does nothing

        Args:
            data (typing.Dict): The data to load

        Returns:
            typing.Dict: The result
        """
        return data

    def dump_data(self, data: typing.Dict) -> typing.Dict:
        """Does not do anything 

        Args:
            data (typing.Any): The data 

        Returns:
            typing.Any: The data
        """
        return data

    def write_text(self, data: typing.Any) -> str:
        """Write the data to a a string

        Args:
            data (typing.Any): The data to write

        Returns:
            str: The string version of the data
        """
        return json.dumps(data)

    def template(self) -> str:
        """Output the template for the class

        Returns:
            str: The template for the output
        """
        return escape_curly_braces(self.key_descr)



# 1st party
from abc import ABC, abstractmethod
import typing
from typing import Self, Any
import itertools
import time
import asyncio
import threading
from enum import Enum
from dataclasses import dataclass
from functools import wraps

# 3rd party
import numpy as np

# local
from ._core import Module
from ._core import Param
from ..utils import UNDEFINED, Renderable
from ._core import render


@dataclass
class Partial(object):
    """Class for storing a partial output from a streaming process
    """
    cur: typing.Any
    prev: typing.Any = None
    dx: typing.Any = None
    complete: bool = False


class I(object):
    """Use to mark data that should not be batched
    """
    
    def __init__(self, data, n: int):
        """Create an I object

        Args:
            data: 
            n (int): The number of times to loop
        """
        self.data = data
        self.n = n

    def __iter__(self) -> typing.Iterator[typing.Any]:
        """Iterate over the object (n times)

        Returns:
            typing.Iterator[typing.Any]: The 
        """
        for _ in range(self.n):
            yield self.data


class P(object):
    """Use to mark data for batching
    """
    
    def __init__(self, data: typing.Iterable, n: int=None):
        """Create a P object that will loop over the data

        Args:
            data (typing.Iterable): The data to loop over
            n (int, optional): The number of items to loop over. Defaults to None.
        """
        self.data = data
        self._n = n or len(data)

    def __iter__(self) -> typing.Iterator:
        """

        Yields:
            typing.Any: Get each value in the  
        """
        for d_i in self.data:
            yield d_i

    @classmethod
    def m(cls, *data: typing.Iterable, n: int=None) -> typing.Tuple['P']:
        """Wrap multiple data through Ps

        data: The data to wrap in P
        n: The number of batches to have

        Returns:
            typing.Tuple[P]: The resulting ps
        """
        return tuple(
            P(d, n) for d in data
        )
    
    @property
    def n(self) -> int:
        """Get the number of iterations

        Returns:
            int: The number of iterations in the loop
        """
        return self._n


def parallel_loop(modules: typing.Union['ModuleList', Module, None], *args, **kwargs) -> typing.Iterator[
    typing.Tuple[Module, typing.List, typing.Dict]
]:
    """Use to loop over the module list

    Args:
        modules (typing.Union[ModuleList, Module, None]): 

    Returns:
        typing.Iterator[ typing.Tuple[Module, typing.List, typing.Dict] ]: 

    Yields:
        Iterator[typing.Iterator[ typing.Tuple[Module, typing.List, typing.Dict] ]]: 
    """
    if isinstance(modules, typing.List):
        modules = ModuleList(modules)

    sz = None
    if isinstance(modules, ModuleList):
        sz = len(modules)
    for arg in itertools.chain(args, kwargs.values()):
        if isinstance(arg, P):
            if sz is None:
                sz = arg.n
            elif arg.n != sz:
                raise ValueError()
    
    if sz is None:
        raise ValueError('None of the inputs can be parallelized')
    
    if not isinstance(modules, ModuleList):
        modules = I(modules, sz)

    kwarg_names = list(kwargs.keys())
    kwarg_vals = [I(v, sz) if not isinstance(v, P) else v for v in kwargs.values()]
    args = [I(v, sz) if not isinstance(v, P) else v for v in args]
    
    for vs in zip(modules, *args, *kwarg_vals):
        m = vs[0]
        args = vs[1:len(args) + 1]
        kwargs = vs[len(args) + 1:]
        yield m, args, dict(zip(kwarg_names, kwargs))


class ParallelModule(Module, ABC):
    """Module that allows 
    """

    def __init__(self, modules: typing.Union[typing.List[Module], Module]):
        """

        Args:
            modules (typing.Union[typing.List[Module], Module]): 
        """
        self._modules = modules

    @abstractmethod
    def forward(self, *args, **kwargs) -> typing.List:
        """Execute the parallel module

        Returns:
            typing.List: The results of the parallel module
        """
        pass


class MultiModule(ParallelModule):
    """A module that executes each of the modules it wraps in a loop
    """
    
    def forward(self, *args, **kwargs) -> typing.List:
        """Loop over all of the modules and execute them

        Returns:
            typing.List: The output of all the modules
        """
        res = []
        for module, cur_args, cur_kwargs in parallel_loop(self._modules, *args, **kwargs):
           res.append(module(*cur_args, **cur_kwargs))
        return res


class AsyncModule(ParallelModule):
    """A type of Parallel module that makes use of 
    Python's Async
    """
    async def async_forward(self, *args, **kwargs) -> typing.List:
        """The asynchronous method to use for inputs

        Args:
            args (typing.List[Args]): The list of inputs to the modules 

        Returns:
            typing.Tuple: The output to the paralell module
        """
        tasks = []
        async with asyncio.TaskGroup() as tg:
            for module, cur_args, cur_kwargs in parallel_loop(self._modules, *args, **kwargs):
                tasks.append(tg.create_task(
                    module.async_forward(*cur_args, **cur_kwargs)
                ))

        return list(
            t.result() for t in tasks
        )

    def forward(self, *args, **kwargs) -> typing.Tuple:
        """Run the asynchronous module
        Returns:
            typing.Tuple: The output for the paralell module
        """
        return asyncio.run(self.async_forward(*args, **kwargs))


# class _ProcessMethod(Module):
#     """
#     """

#     def __init__(self, f: typing.Callable, instance=None):
#         self.f = f
#         self.instance = instance
#         self._stored = None
#         self._async_f = None

#     def forward(self, *args, **kwargs) -> typing.Any:
#         if self.instance:
#             return self.f(self.instance, *args, **kwargs)
#         return self.f(*args, **kwargs)

#     async def async_forward(self, *args, **kwargs) -> typing.Any:
        
#         if self._async_f:
#             self._async_f(*args, **kwargs)
#         return self.forward(*args, **kwargs)

#     def __get__(self, instance, owner):

#         if self._stored is not None and instance is self._stored:
#             return self._stored
#         self._stored = _ProcessMethod(self.f, instance)
#         return self._stored
    
#     @classmethod
#     def async_(cls, f):

#         cls._async_f = f

# TODO: Decide whether to remove this
# def processfunc(f):
#     """Decorator to create a module from a method

#     Args:
#         f (_type_): _description_

#     Returns:
#         _type_: _description_
#     """

#     @wraps(f)
#     def wrapper(*args, **kwargs):
#         return f(*args, **kwargs)

#     if hasattr(f, '__self__') or '__self__' in dir(f):
#         return _ProcessMethod(f)
#     else:
#         return _ProcessMethod(wrapper)


class ModuleList(Module):
    """Use to have a list of modules
    """

    def __init__(self, modules: typing.List[Module]):
        """Create a list of modules

        Args:
            modules (typing.List[Module]): the modules to set
        """
        self._modules = modules

    def children(self, recurse: bool = True) -> typing.Iterator[Module]:
        """Get the children of the ModuleList

        Args:
            recurse (bool, optional): Whether to recurse to child modules. Defaults to True.

        Yields:
            Module: The children of the module
        """
        for module in self._modules:
            yield module
            if recurse:
                for child in module.children(recurse):
                    yield child

    def parameters(self, recurse: bool = True) -> typing.Iterator[Param]:
        """Get the parameters in the ModuleList

        Args:
            recurse (bool, optional): _description_. Defaults to True

        Yields:
            Param: The params in the ModuleList
        """
        for module in self._modules:
            for p in module.parameters(recurse):
                yield p
        
    def forward(self) -> Any:
        """

        Raises:
            ValueError: If called because ModuleList cannot be __call__

        """
        raise ValueError('Cannot pass forward with ModuleList')
    
    def __len__(self) -> int:
        """Get the number of modules in the ModuleList

        Returns:
            int: The number of modules
        """
        return len(self._modules)
    
    def __iter__(self) -> typing.Iterator[Module]:
        """Loop over the modules in the ModuleList

        Yields:
            Module: All of the modules in the ModuleList
        """
        for m in self._modules:
            yield m


class Sequential(ModuleList):
    """Use to 
    """
    
    def __init__(self, *module) -> None:
        """Create a Sequence of modules
        """
        super().__init__(module)

    def add(self, module):
        """Add a module to Sequential

        Args:
            module: Add the module to the Sequential
        """
        self._modules.append(module)

    def forward(self, *x) -> Any:
        """Pass the input (x) through each of the modules

        Returns:
            Any: The result of the final model
        """
        
        multi = len(x) > 1
        if len(x) == 1:
            x = x[0]

        first = True
        for module in self._modules:
            if first and multi:
                x = module(*x)
            else:
                x = module(x)
            first = False
        return x


class Batched(Renderable):
    """Batched is data 
    """

    def __init__(
        self, *data: typing.Iterable, 
        size: int=1, drop_last: bool=True, 
        order: typing.Optional[typing.List]=None
    ):
        """Create Batched data

        Args:
            size (int, optional): the size of each batch. Defaults to 1.
            drop_last (bool, optional): Whether to drop the last element. Defaults to True.
            order (typing.Optional[typing.List], optional): The order to return the values in. Defaults to None.
        """
        if len(data) == 0:
            raise ValueError('No data was passed in to batch')
        sz = None
        for v in data:
            if sz is None:
                sz = len(v)
            elif sz != len(v):
                raise ValueError(
                    'The lengths of all of the elements '
                    'to batch must be the same'
                )
        self._data = data
        self._size = size
        self._n_elements = sz
        self.drop_last = drop_last
        add_one = (self._n_elements % self._size) != 0
        self._n = (self._n_elements // self._size) + add_one
        self._order = (
            order if order is not None 
            else np.arange(self._n_elements)
        )

    def __len__(self) -> int:
        """Get the length of the batch

        Returns:
            int: The number of elements in the batch
        """
        return self._n

    def shuffle(self) -> Self:
        """Shuffle the batched data

        Returns:
            Self: Batched but shuffled
        """
        return Batched(
            *self._data, size=self._size, drop_last=self.drop_last, 
            order=np.random.permutation(self._n_elements)
        )

    def __iter__(self) -> typing.Iterator[typing.Any]:
        """Iterate over the batched data

        Yields:
            Iterator[typing.Iterator[typing.Any]]: _description_
        """
        start = 0
        upto = self._size
        for _ in range(self._n):
            cur = []
            for data in self._data:
                if isinstance(data, np.ndarray):
                    cur_data = data[self._order[start:upto]]
                else:
                    cur_data = [data[i] for i in self._order[start:upto]]
                cur.append(cur_data)
    
            start = upto
            upto += self._size
            if len(cur) == 1:
                yield cur[0]
            else:
                yield tuple(cur)
    
    def render(self) -> str:
        """Render the batch to a string

        Returns:
            str: The text
        """
        return '\n'.join(
            f'{i}: {render(v)}' for i, v in enumerate(self)
        )


def reduce(mod: Module, *args, init_mod: Module=None, init_val=None, **kwargs):
    """Reduce the args passed in with a module

    Args:
        mod (Module): The module to use for reduction
        init_mod (Module, optional): The module to use for the first set of data. Defaults to None.
        init_val: The initial value to use
        
    Returns:
        The result of the reduction
    """
    results = []

    for _, cur_args, cur_kwargs in parallel_loop(None, *args, **kwargs):

        if len(results) == 0 and init_mod is not None and init_val is None:
            results.append(init_mod(*cur_args, **cur_kwargs))
        elif len(results) == 0 and init_mod is not None:
            results.append(init_mod(init_val, *cur_args, **cur_kwargs))
        elif len(results) == 0:
            results.append(mod(init_val, *cur_args, **cur_kwargs))
        else:
            results.append(mod(results[-1], *cur_args, **cur_kwargs))
    return results[-1]


async def _async_map(f: Module, *args, **kwargs) -> typing.Tuple[typing.Any]:
    """Helper function to run async_map

    Args:
        f (Module): The function to asynchronously execute

    Returns:
        typing.Tuple[typing.Any]: The result of the map
    """
    tasks = []
    async with asyncio.TaskGroup() as tg:
        
        for cur_f, a, kv in parallel_loop(f, *args, **kwargs):

            tasks.append(tg.create_task(cur_f.async_forward(*a, **kv)))

    return tuple(
        t.result() for t in tasks
    )


def async_map(f: Module, *args, **kwargs) -> typing.Tuple[typing.Any]:
    """Map *args and **kwargs through the module

    Args:
        f (Module): The module to use for mapping

    Returns:
        typing.Tuple[typing.Any]: The output of the modules
    """
    return asyncio.run(_async_map(f, *args, **kwargs))


async def _async_multi(*f) -> typing.Tuple[typing.Any]:
    """Helper function to run asynchronous processes

    Returns:
        typing.Tuple[typing.Any]: 
    """

    tasks = []
    async with asyncio.TaskGroup() as tg:

        for f_i in f:
            tasks.append(tg.create_task(f_i))

    return tuple(
        task.result() for task in tasks
    )


def async_multi(*f) -> typing.Tuple[typing.Any]:
    """Run multiple asynchronous processes and return
    the results

    Args:
       f (coroutine): The coroutines to run

    Returns:
        typing.Tuple: The output of the 
    """
    return asyncio.run(_async_multi(*f))


@dataclass
class Partial(object):
    """Class for storing a partial output from a streaming process
    """
    cur: typing.Any
    prev: typing.Any = None
    dx: typing.Any = None
    complete: bool = False


class Streamer(object):
    """Streamer is an object used to stream over the response
    """

    def __init__(self, stream: typing.Iterable):
        """The Stream to loop over

        Args:
            stream: The stream to loop over in generating the stream
        """
        self._stream = iter(stream)
        self._cur = None
        self._output = UNDEFINED
        self._prev = None
        self._dx = None

    @property
    def complete(self) -> bool:
        """Check if the stream has completed

        Returns:
            bool: True if the stream is complete
        """
        return self._output is not UNDEFINED

    def __call__(self) -> typing.Union[Partial]:
        """Query the streamer and returned updated value if updated

        Returns:
            typing.Union[typing.Any, Partial]: Get the next value in the stream
        """
        if self._output is not UNDEFINED:
            return self._output
        try:
            self._prev = self._cur
            self._cur, self._dx = next(self._stream)
            return Partial(self._cur, self._prev, self._dx, False)    
        except StopIteration:
            self._output = Partial(self._cur, self._prev, self._dx, True) 
            return self._output
        
    def __iter__(self) -> typing.Iterator[Partial]:
        """Iterate over the stream

        Yields:
            Partial: The partial results of the stream
        """
        while True:

            cur = self()
            if cur.complete:
                break
            yield cur


class RunStatus(Enum):
    """The status for a "Run"
    """

    RUNNING = 'running'
    READY = 'ready'
    FINISHED = 'finished'


class Runner(object):
    """Create an object to loop over the stream
    """

    def __init__(self, module, *args, **kwargs):
        """Create a runner which will loop over the stream

        Args:
            module: The module to stream
        """
        
        self.module = module
        self.status = RunStatus.READY
        args = [module, *args]
        self._result = None
        self.t = threading.Thread(target=self._exec, args=args, kwargs=kwargs)
        self.t.run()

    @property
    def result(self) -> typing.Any:
        """Return the result

        Returns:
            typing.Any: The result of the 
        """
        return self._result

    def _exec(self, module: Module, *args, **kwargs):
        """The method that gets executed by the thread

        Args:
            module (Module): The module to execute
        """
        self.status = RunStatus.RUNNING
        result = module(*args, **kwargs)
        self.status = RunStatus.FINISHED
        self._result = result


class StreamRunner(object):
    """Use to run a stream
    """

    def __init__(self, module,  *args, **kwargs):
        """Create a StreamRunner which will run the module in 
        a stream

        Args:
            module: The module to stream
        """
        self.module = module
        self.status = RunStatus.READY
        args = [module, *args]
        self._results = []
        self._result_dx = []
        self.t = threading.Thread(target=self._stream_exec, args=args, kwargs=kwargs)
        self.t.run()

    def _stream_exec(self, module: Module, *args, **kwargs):
        """Stream the module

        Args:
            module (Module): The module to run
        """

        self.status = RunStatus.RUNNING
        for d, dx in module.stream_forward(*args, **kwargs):
            self._result_dx.append(dx)
            self._results.append(d)

        self.status = RunStatus.FINISHED

    @property
    def result(self) -> typing.Any:
        """Get the result of the stream

        Returns:
            typing.Any: The result
        """
        if len(self._results) > 0:
            return self._results[-1]
        return None

    @property
    def result_dx(self) -> typing.Any:
        """Get the result of the stream

        Returns:
            typing.Any: The result
        """
        if len(self._result_dx) > 0:
            return [*self._result_dx]
        return None

    def dx(self) -> typing.Iterator:
        """Loop over the changes in the results

        Yields:
            typing.Any: Each change" in value
        """
        for x in self._result_dx:
            yield x

    def exec_loop(self, sleep_time: float=1./60) -> typing.Iterator[typing.Tuple]:
        """Execute the loop

        Args:
            sleep_time (float, optional): The interval to execute the loop at. Defaults to 1./60.

        Yields:
            typing.Tuple: Each result and dx
        """
        i = 0
        while True:
            if len(self._results) > i:
                yield self._results[i], self._result_dx[i]
                i += 1
            
            elif self.status == RunStatus.FINISHED:
                break
            
            time.sleep(sleep_time)

    def __iter__(self)  -> typing.Iterator[typing.Tuple]:
        """Loop over the results and the dxs

        Yields:
            typing.Tuple: The result and each change in the result
        """
        for r, rx in zip(self._results, self._result_dx):
            yield r, rx


def run_thread(module: Module, *args, **kwargs) -> Runner:
    """Convenience function to create a "Runner"

    Args:
        module (Module): The module to execute

    Returns:
        Runner: 
    """
    return Runner(
        module, *args, **kwargs
    )


def stream_thread(module, *args, **kwargs) -> StreamRunner:
    """Convenience function to stream a thread. Will create a StreamRunner

    Args:
        module: The module to run

    Returns:
        StreamRunner: The stream runner
    """
    return StreamRunner(
        module, *args, **kwargs
    )



from ._core import (
    Reader, NullRead, 
    Storable, render,
    render_multi, 
    Module, Cue,
    Param

)
from ._read import (
    MultiRead, PrimRead, PydanticRead
)
from ._ai import (
    AIModel, AIPrompt, AIResponse, Dialog, Message,
    TextMessage, Data,
    stream_text
)
from ._process import (
    Partial, 
    ParallelModule, 
    parallel_loop,
    MultiModule, 
    ModuleList,
    Sequential, 
    Batched,
    Streamer, 
    AsyncModule,
    async_multi,
    reduce,
    I,
    P,
    async_map,
    run_thread,
    stream_thread,
    Runner,
    RunStatus,
    StreamRunner
)
from ._instruct import (
    validate_out, 
    InstructCall, SignatureFunc, signaturefunc,
    signaturemethod, instructfunc,
    InstructFunc, instructmethod,

)



import typing
import json

import pydantic

from ..utils import (
    struct_template, model_to_text, 
    unescape_curly_braces, TemplateField, 
    StructLoadException
)
from ._core import Reader, render

from pydantic_core import PydanticUndefined


S = typing.TypeVar('S', bound=pydantic.BaseModel)


class MultiRead(Reader):
    """Concatenates multiple outputs Out into one output
    """
    outs: typing.List[Reader]
    conn: str = '::OUT::{name}::\n'
    signal: str = '\u241E'

    def dump_data(self, data: typing.Dict) -> typing.Dict:
        """Dump all of the data to a dictionary

        Args:
            data (typing.Dict): A dictionary containing all
            of the data

        Returns:
            typing.Dict: The data all dumped
        """
        results = []
        for data_i, out in zip(data['data'], self.outs):
            results.append(out.dump_data(data_i))
        return {'data': results, 'i': data['i']}

    def write_text(self, data: typing.Dict) -> str:
        """Convert the dictionary data to text

        Args:
            data (typing.Dict): The dictionary data obtained from dump_data

        Returns:
            str: The data written as text
        """
        result = ''
        for i, (d, out) in enumerate(zip(data['data'], self.outs)):
            name = out.name or str(i) 
            result = result + '\n' + self.signal + self.conn.format(name=name)
            result = f'{result}\n{render(d)}'

        return result

    def to_text(self, data: typing.List[S]) -> str:
        """Convert the data to text
        """

        text = ""
        for i, (data_i, out) in enumerate(zip(data, self.outs)):
            name = out.name or str(i) 
            cur = render(data_i)
            cur_conn = self.conn.format(name)
            text += f"""
            {self.signal}{cur_conn}
            {cur}
            """
        return text

    def read_text(self, message: str) -> typing.Dict:
        """

        Args:
            message (str): _description_

        Returns:
            typing.Dict: _description_
        """
        structs = []

        d = message
        for i, out in enumerate(self.outs):
            name = out.name or str(i)
            from_loc = d.find('\u241E')
            to_loc = d[from_loc + 1:].find('\u241E')
            cur = self.conn.format(name=name)
            data_loc = from_loc + len(cur) + 1

            if to_loc != -1:
                data_end_loc = from_loc + 1 + to_loc
            else:
                data_end_loc = None

            data_str = d[data_loc:data_end_loc]
            try: 
                structs.append(out.read_text(data_str))
            except StructLoadException as e:
                return {'data': structs, 'i': i}
            d = d[data_end_loc:]

        return {'data': structs, 'i': i, 'n': len(self.outs)}
    
    def load_data(self, data: typing.Dict) -> typing.Dict:
        """Load the data for each of the components of the output

        Args:
            data (): The dictionary data to load

        Returns:
            typing.Dict: The data with each element of the output
             loaded
        """
        structs = []

        for o, d_i in zip(self.outs, data['data']):
            structs.append(o.load_data(d_i))

        return {'data': structs, 'i': data['i'], 'n': data['n']}

    def template(self) -> str:
        """Output a template for the output

        Args:
            data (): The dictionary data to load

        Returns:
            typing.Dict: The data with each element of the output
             loaded
        """
        texts = []
        for i, out in enumerate(self.outs):
            print(out.name)
            name = out.name or str(i)
            cur = struct_template(out)
            print(out, type(out))
            cur_conn = self.conn.format(name=name)
            texts.append(f"{self.signal}{cur_conn}\n{cur}")
        text = '\n'.join(texts)
        return text


class PrimRead(Reader):
    """Use for converting an AI response into a primitive value
    """

    _out_cls: typing.Type

    def __init__(
        self, out_cls: typing.Type,
        **data
    ):
        """Create a reader for Primitive values

        Args:
            out (typing.Type): The type of data
        """
        super().__init__(**data)
        self._out_cls = out_cls

    def to_text(self, data: typing.Any) -> str:
        """Convert the data to text

        Args:
            data (typing.Any): The data to convert

        Returns:
            str: The text
        """
        return str(data)

    def dump_data(self, data: typing.Any) -> typing.Any:
        """

        Args:
            data (typing.Any): The data to dump

        Returns:
            typing.Any: The data (since it is a primitive, does nothing)
        """
        return data

    def write_text(self, data: typing.Any) -> str:
        """Convert the primitive to a string

        Args:
            data (primitive): The primitive

        Returns:
            str: The data converted to a string
        """
        return str(data)

    def read_text(self, message: str) -> typing.Any:
        """Read the message from text

        Args:
            message (str): The message to read

        Returns:
            typing.Any: The string converted to the primitive
        """
        if self._out_cls is bool:
            return message.lower() in ('true', 'y', 'yes', '1', 't')
        return self._out_cls(message)
    
    def load_data(self, data) -> typing.Any:
        """Doesn't do anything because the data should be in
        the right form

        Args:
            data: The primitive

        Returns:
            typing.Any: The primitive value
        """
        return data

    def template(self) -> str:
        """Output the template for the string

        Returns:
            str: The template for the data
        """
        return f'<{self._out_cls}>'


class PydanticRead(Reader, typing.Generic[S]):
    """Use for converting an AI response into a Pydantic BaseModel
    """
    _out_cls: typing.Type[S] = pydantic.PrivateAttr()

    def __init__(self, out_cls: S, **data):
        """Read in a 

        Args:
            out_cls (S): The class to read in
        """
        super().__init__(**data)
        self._out_cls = out_cls

    def dump_data(self, data: S) -> typing.Any:
        """Convert the data to a dictionary

        Args:
            data (S): The data to convert

        Returns:
            typing.Any: The result
        """
        return data.model_dump()

    def write_text(self, data: typing.Any) -> str:
        """Convert the dumped data to a string

        Args:
            data (typing.Any): The data to convert to a string

        Returns:
            str: The convered data
        """
        return str(data)

    def to_text(self, data: S) -> str:
        """Convert the data to text

        Args:
            data (S): The data to convert

        Returns:
            str: The 
        """
        return model_to_text(data, True)

    def read_text(self, message: str) -> S:
        """Read in text from the message

        Args:
            message (str): Read in the text from 

        Returns:
            S: The BaseModel specified by S
        """
        message = unescape_curly_braces(message)
        return json.loads(message)
        # return self._out_cls.from_text(message, True)
    
    def load_data(self, data) -> S:
        """Load data from the 

        Args:
            data: The data to load

        Returns:
            S: the loaded data
        """
        return self._out_cls(**data)
    
    def template_renderer(self, template: TemplateField) -> str:
        """Render the template for the BaseModel

        Args:
            template (TemplateField): The template to render

        Returns:
            str: The template
        """
        t = f'<{template.description}> - type: {template.type_}'
        if template.default is not None or template.default == PydanticUndefined:
            t = f'{t} {template.default}>'
        else:
            t = f'{t}>'
        return t

    def template(self, escape_braces: bool=False) -> str:
        """Convert the object to a template

        Args:
            escape_braces (bool, optional): Whether to escape curly brances. Defaults to False.

        Returns:
            str: Escape the braces
        """
        # return self._out_cls.template()
        return render(struct_template(self._out_cls), escape_braces, self.template_renderer) 




# 1st party
import typing
from typing import Self
import typing

from ..utils import Renderable

# 3rd party
import pydantic


class Term(pydantic.BaseModel, Renderable):
    """Use to define a term used in the AI system
    for repeated usage
    """

    name: str
    definition: str
    meta: typing.Dict[str, typing.Union[str, typing.List[str]]]

    def __init__(self, name: str, definition: str, **meta: str):

        super().__init__(
            name=name, definition=definition, meta=meta
        )

    def __getitem__(
        self, key: typing.Union[str, typing.Iterable[str]]
    ) -> typing.Union[str, typing.List[str]]:
        """Get one or more items from meta

        Args:
            key (typing.Union[str, typing.Iterable[str]]): The key to retrieve with

        Returns:
            typing.Union[str, typing.List[str]]: The value retrieved by key
        """
        if isinstance(key, str):
            return self.meta[key]
        
        return [
            v for k, v in self.meta.items() if k in key
        ]

    def __setitem__(self, key: typing.Union[str, typing.List[str]], val: typing.Union[str, typing.List[str]]):
        """Set the meta data for the term

        Args:
            key (typing.Union[str, typing.List[str]]): the key or keys to the meta data
            val (typing.Union[str, typing.List[str]]): the data to set

        Returns:
            val
        """
        if isinstance(key, str):
            self.meta[key] = val
        else:
            for k, v in zip(key, val):
                self.meta[k] = v
        return val
    
    def render(self) -> str:
        """Render the DataList

        Returns:
            str: The data list rendered as a string
        """
        base = f"{self.name}: {self.definition}"
        if len(self.meta) == 0:
            return base
        
        meta = '\n-'.join(f'{k}: {v}' for k, v in self.meta.items())
        return f'{base}\n{meta}'


class Glossary(pydantic.BaseModel, Renderable):
    """A glossary contains a list of terms
    """

    terms: typing.Dict[str, Term]

    def __init__(self, terms: typing.List[Term] = None):
        """Create a glossary of terms

        Args:
            terms (typing.List[Term], optional): . Defaults to None.
        """
        terms = terms or []
        super().__init__(
            terms={term.name: term for term in terms}
        )

    def __getitem__(
        self, key: typing.Union[str, typing.Iterable[str]]
    ) -> typing.Union[str, typing.List[str]]:
        """Get an item from the glossary

        Args:
            key (typing.Union[str, typing.Iterable[str]]): 

        Returns:
            typing.Union[str, typing.List[str]]: 
        """
        if isinstance(key, str):
            return self.terms[key]
        
        return [
            v for k, v in self.terms.items() if k in key
        ]

    def join(
        self, val: typing.Union[Term, typing.List[Term]]
    ):
        """Join the glossary with terms

        Args:
            val (typing.Union[Term, typing.List[Term]]): The value to add

        Returns: The value joined
            ...
        """
        if isinstance(val, Term):
            self.terms[val.name] = val
        else:
            for term_i in val:
                self.terms[term_i.name] = val
        return val
    
    def add(self, name: str, definition: str, **meta) -> Self:
        """Add a term to the glossary

        Args:
            name (str): The name of the term to add
            definition (str): The definition of the term

        Returns:
            Self: The glossary
        """
        self.terms[name] = Term(name, definition, **meta)
        return self
    
    def exclude(self, *meta: str) -> 'Glossary':
        """Filter by excluding meta fields from the glossary

        Returns:
            Glossary: The filtered glossary
        """
        terms = []
        for name, term in self.terms.items():
            cur_meta = {
                k: v for k, v in term.meta.items() if k not in meta
            }
            terms.append(Term(name, **cur_meta))
        return Glossary(terms)

    def include(self, *meta: str) -> 'Glossary':
        """Filter by including meta fields from the glossary

        Returns:
            Glossary: The filtered glossary
        """

        terms = []
        for name, term in self.terms.items():
            cur_meta = {
                k: v for k, v in term.meta.items() if k in meta
            }
            terms.append(Term(name, **cur_meta))
        return Glossary(terms)

    def render(self) -> str:
        """Convert the glossary into a string

        Returns:
            str: The name of the string
        """
        return '\n'.join(term.render() for _, term in self.terms.items())

    def items(self) -> typing.Iterator[typing.Tuple[str, Term]]:
        """Looop over the items in the Glossary

        Returns:
            typing.Iterator[typing.Tuple[str, Term]]: The iterator for the terms
        """
        return self.terms.items()



from ._tasks import TaskStatus, Task
import threading
import typing

import time
from enum import Enum
from . import Root


class AgentStatus(Enum):

    READY = "ready"
    PAUSED = "paused"
    RUNNING = "running"
    STOPPED = "stopped"
    WAITING = "waiting"


class Agent(object):
    """Agent class
    """

    def __init__(
        self, sango: Root=None, interval: float=None
    ):
        """

        Args:
            sango (Sango, optional): The default sango to use if act() is not overridden. Defaults to None.
            interval (float, optional): The interval to repeat actions at. Defaults to None.
        """
        super().__init__()
        self._task_status = None
        self._status = None
        self._sango = sango
        self._interval = interval
        self._queued = []

    def act(self) -> TaskStatus:
        """Execute the agent

        Returns:
            TaskStatus: The resulting status
        """
        if self._sango is None:
            return TaskStatus.SUCCESS
        return self._sango.tick()

    def pause(self):
        self._status = AgentStatus.PAUSED

    def reset(self):
        self._status = AgentStatus.READY

    def stop(self):
        self._status = AgentStatus.STOPPED

    def run(self, continue_on_fail: bool) -> typing.Iterator[TaskStatus]:
        """Execute the agent at the specified interval
        """
        while self._status != AgentStatus.STOPPED:
            self._task_status = self.act()
            if self._interval is not None:
                time.sleep(self._interval)
            if (
                (not continue_on_fail and self._task_status == TaskStatus.FAILURE)
            ):
                self._status = AgentStatus.STOPPED
            yield self._task_status



# 1st party
from abc import ABC, abstractmethod
from typing import Self
import typing
from ..utils import (
    Renderable, TemplateField, 
    Templatable, model_to_text
)
from uuid import uuid4

# 3rd party
import pydantic

# local
from ..utils import (
    is_primitive, 
    escape_curly_braces
)

S = typing.TypeVar('S', bound=pydantic.BaseModel)


# TODO: Make "struct" storable?
#  Module as well
# Consider whether to remove

class Storable(ABC):
    """Object to serialize objects to make them easy to recover
    """
    def __init__(self):
        """Create the storable object
        """
        self._id = str(uuid4())

    @property
    def id(self) -> str:
        """The object id of the storable

        Returns:
            str: The ID
        """
        return self._id

    def load_state_dict(self, state_dict: typing.Dict):
        """Load the state dict for the object

        Args:
            state_dict (typing.Dict): The state dict
        """
        for k, v in self.__dict__.items():
            if isinstance(v, Storable):
                self.__dict__[k] = v.load_state_dict(state_dict[k])
            else:
                self.__dict__[k] = state_dict[k]
        
    def state_dict(self) -> typing.Dict:
        """Retrieve the state dict for the object

        Returns:
            typing.Dict: The state dict
        """
        cur = {}

        for k, v in self.__dict__.items():
            if isinstance(v, Storable):
                cur[k] = v.state_dict()
            else:
                cur[k] = v
        return cur


def render(
    x: typing.Any, escape_braces: bool=True, 
    template_render: typing.Optional[typing.Callable[[TemplateField], str]]=None
) -> typing.Union[str, typing.List[str]]:
    """Convert an input to text. Will use the text for a cue,
    the render() method for a description and convert any other value to
    text with str()

    Args:
        value (X): The input

    Returns:
        str: The resulting text
    """
    if isinstance(x, TemplateField):
        if template_render is not None:
            x = template_render(x)
        else: 
            x = x.render()

    if isinstance(x, Renderable):
        return x.render()

    elif isinstance(x, pydantic.BaseModel):
        return model_to_text(x, escape_curly_braces)
    elif is_primitive(x):
        return str(x)
    elif isinstance(x, typing.Dict):
        items = {}
        for k, v in x.items():
            v = render(v, escape_braces)
            if isinstance(v, str):
                v = f'"{v}"'
            items[k] = v    
        items = ', '.join(
            f'"{k}": {v}' 
            for k, v in items.items()
        )

        if escape_braces:
            return f"{{{{{items}}}}}"
        else:
            return f'{{{items}}}'
    elif isinstance(x, typing.List):

        items = []
        for v in x:
            v = render(v, escape_braces)
            if isinstance(v, str):
                v = f'"{v}"'
            items.append(v)

        return '[{}]'.format(', '.join(render(v) for v in items))
    elif isinstance(x, Renderable):
        return x.render()
    return str(x)


def render_multi(xs: typing.Iterable[typing.Any]) -> typing.List[str]:
    """Convert an input to text. Will use the text for an cue,
    the render() method for a description and convert any other value to
    text with str()

    Args:
        value (X): The input

    Returns:
        str: The resulting text
    """
    return [
        render(x) for x in xs
    ]


class Reader(pydantic.BaseModel, Templatable, ABC):
    """Use a reader to read in data convert data retrieved from
    an LLM to a better format
    """

    name: str = ''

    def example(self, data: typing.Any) -> str:
        """Output an example of the data

        Args:
            data (typing.Any): 

        Returns:
            str: 
        """
        return self.write_text(self.dump_data(data))

    @abstractmethod
    def dump_data(self, data: typing.Any) -> typing.Any:
        """Convert the data from the output of write_text
        to the original format

        Args:
            data (typing.Any): The data

        Returns:
            typing.Any: The data
        """
        pass

    @abstractmethod
    def write_text(self, data: typing.Any) -> str:
        """Write out the text for the data

        Args:
            data (typing.Any): The data to write the text for

        Returns:
            str: The text
        """
        pass

    def read(self, message: str) -> typing.Any:
        """Read in the output

        Args:
            message (str): The message to read

        Returns:
            typing.Any: The output of the reader
        """
        return self.load_data(self.read_text(message))
    
    @abstractmethod
    def read_text(self, message: str) -> typing.Any:
        """Read in the text and output to a "json" compatible format or a primitive

        Args:
            message (str): The message to read

        Returns:
            typing.Any: The result of the reading process
        """
        pass

    @abstractmethod
    def load_data(self, data: typing.Dict) -> typing.Any:
        """Load the data output from reading the text

        Args:
            data (typing.Dict): The data to load (JSON format)

        Returns:
            typing.Any: The result of the reading
        """
        pass

    @abstractmethod
    def template(self) -> str:
        """Get the template for the reader

        Returns:
            str: The template as a string
        """
        pass


class NullRead(Reader):
    """A Reader that does not change the data. 
    So in most cases will simply output a string
    """

    def dump_data(self, data: typing.Any) -> typing.Any:
        """Convert the data to JSON compatible format

        Args:
            data (typing.Any): The data to convert to "JSON" compatible format

        Returns:
            typing.Any: Returns the data passed in
        """
        return data

    def write_text(self, data: typing.Any) -> str:
        """Output the data to text

        Args:
            data (typing.Any): The JSON compatible data

        Returns:
            str: The data converted to text
        """
        return data

    def read_text(self, data: str) -> typing.Dict:
        """Read in the text as a JSON compatible structure

        Args:
            data (str): The data to read in

        Returns:
            typing.Dict: The JSON compatible object (does nothing because it is null)
        """
        return data
    
    def load_data(self, data) -> typing.Any:
        """Load the data

        Args:
            data: The data to load

        Returns:
            typing.Any: The data passed in (since null)
        """
        return data

    def template(self) -> str:
        return None


class Instruct(ABC):
    """
    """
    @abstractmethod
    def i(self) -> 'Cue':
        """Create an Instruct class used for instructions

        Returns:
            Cue: Get the cue
        """
        pass


class Cue(pydantic.BaseModel, Instruct, typing.Generic[S], Renderable):
    """Specific cue for the model to use
    """
    
    text: str
    out: typing.Optional[Reader] = None

    def __init__(self, text: str, name: str='', out: typing.Optional[Reader] = None):

        super().__init__(text=text, name=name, out=out)

    def i(self) -> Self:
        return self

    @pydantic.field_validator('text', mode='before')
    def convert_renderable_to_string(cls, v):
        if isinstance(v, Renderable):
            return v.render()
        if is_primitive(v):
            return str(v)
        return v

    def render(self) -> str:
        """Render the cue

        Returns:
            str: The text for the cue 
        """
        return self.text

    def read(self, data: str) -> S:
        """Read the data

        Args:
            data (str): The data to read

        Raises:
            RuntimeError: If the cue does not have a reader

        Returns:
            S: The result of the read process
        """
        if self.out is None:
            raise RuntimeError(
                "Out has not been specified so can't read it"
            )
        
        return self.out.read(data)

    def state_dict(self) -> typing.Dict:
        
        return {
            'text': self.text,
        }

    def load_state_dict(self, params: typing.Dict):
        
        self.text = params['text']


class Param(pydantic.BaseModel, Renderable, Storable):
    """Use Param to wrap instructions so the instructions
    can update
    """
    
    name: str
    cue: Cue
    training: bool=False
    text: str = None

    @pydantic.field_validator('cue', mode='before')
    def convert_renderable_to_string(cls, v):
        if isinstance(v, Cue):
            return v
        if isinstance(v, Renderable):
            return Cue(text=v.render())
        if is_primitive(v):
            return Cue(text=render(v))
        return v

    def update(self, text: str) -> bool:
        """Update the text for the parameter
        If not in "training" mode will not update

        Args:
            text (str): The text to update with
        
        Returns:
            True if updated and Fals if not (not in training mode)
        """
        if self.training:
            self.text = text
            return True
        return False

    def render(self) -> str:
        """Convert the Parameter to a string
        IF the text for the paramter has not been 
        updated 

        Returns:
            str: 
        """
        if self.text is None:
            return self.cue.render()
        return self.text

    def read(self, data: typing.Dict) -> S:
        """Read in the data

        Args:
            data (typing.Dict): The data to read in

        Returns:
            S: The result of the reading
        """
        return self.cue.read(data)

    def reads(self, data: str) -> S:
        return self.cue.read_out(data)
    
    def state_dict(self) -> typing.Dict:
        """Get the state dict for the Param

        Returns:
            typing.Dict: the state dict
        """
        
        return {
            'name': self.name,
            'cue': self.cue.state_dict(),
            'training': self.training,
            'text': self.text
        }

    def load_state_dict(self, params: typing.Dict):
        """Load the state dict for the Param
        
        Args:
            params (typing.Dict): the state dict
        """
        self.name = params['name']
        self.cue = self.cue.load_state_dict(params['cue'])
        self.training = params['training']
        self.text = params['text']


class Module(Storable, ABC):
    """Base class for Modules
    """

    @abstractmethod
    def forward(self, *args, **kwargs) -> typing.Any:
        """Execute the module

        Returns:
            typing.Any: The output of the module
        """
        pass

    def __call__(self, *args, **kwargs) -> typing.Any:
        """Execute the module

        Returns:
            typing.Any: The output of the module
        """
        return self.forward(*args, **kwargs)

    def parameters(self, recurse: bool=True) -> typing.Iterator['Param']:
        """Loop over the parameters for the module

        Yields:
            Param: The parameters for the module
        """
        yielded = set()
        for k, v in self.__dict__.items():
            if isinstance(v, Param):
                if id(v) in yielded:
                    continue
                yielded.add(id(v))
                
                yield v
            if recurse and isinstance(v, Module):
                for v in v.parameters(True):
                    if id(v) in yielded:
                        continue
                    yielded.add(id(v))
                    yield v

    def children(self, recurse: bool=True) -> typing.Iterator['Module']:
        """Loop over all of the child modules

        Yields:
            Module: The child module
        """
        yielded = set()
        print(self.__dict__)
        for k, v in self.__dict__.items():
            if isinstance(v, Module):
                if id(v) in yielded:
                    continue
                yield v
                yielded.add(id(v))
                if recurse:
                    for v in v.children(True):
                        if id(v) in yielded:
                            continue
                        yielded.add(id(v))
                        yield v
    
    async def async_forward(self, *args, **kwargs) -> typing.Any:
        """Execute the forward method asynchronously

        Returns:
            typing.Any: 
        """
        res = self.forward(*args, **kwargs)
        return res

    def stream_forward(self, *args, **kwargs) -> typing.Iterator[
        typing.Tuple[typing.Any, typing.Any]
    ]:
        """Stream the output

        Yields:
            Iterator[typing.Iterator[ typing.Tuple[typing.Any, typing.Any] ]]: The current value and the change in the value
        """
        # default behavior doesn't actually stream
        res = self.forward(*args, **kwargs) 
        yield res, res

    async def async_stream_forward(self, *args, **kwargs) -> typing.AsyncIterator:
        """
        Returns:
            Streamer: The Streamer to loop over
        """

        for d, dx in self.stream_forward(*args, **kwargs):
            yield d, dx

    def state_dict(self):
        
        state_dict = {}
        for i, child in enumerate(self.children(False)):
            state_dict[i] = child.state_dict()
        
        params = {}
        for i, param in enumerate(self.parameters(False)):
            params[i] = param.state_dict()
        state_dict['__params__'] = params

        return state_dict
    
    def load_state_dict(self, state_dict):

        for i, child in enumerate(self.children(False)):
            cur_dict = state_dict[i]
            child.load_state_dict(cur_dict)

        params = state_dict['__params__']
        for i, cur in enumerate(self.parameters(False)):
            cur.load_state_dict(params[i])



# 1st party
import typing

# 3rd party
import pydantic

# local
from ..utils._utils import (
    generic_class
)
from ..utils._utils import generic_class
from .._core._ai import Message


S = typing.TypeVar('S', bound=pydantic.BaseModel)


class Media:
    """Use to store media
    """
    descr: str
    data: str


class DataList(pydantic.BaseModel, typing.Generic[S]):
    """Create a list of data that inherit from pydantic.BaseModel
    """
    data: typing.List[S]

    def __getitem__(self, key) -> typing.Any:
        """Get a value in the data list

        Args:
            key: The index for the data

        Returns:
            typing.Any: The value at that index
        """
        return self.data[key]
    
    def __setitem__(self, key: typing.Optional[int], value: S) -> typing.Any:
        """Set a value in the 

        Args:
            key (str): The key for the value to set
            value : The value to set

        Returns:
            S: the value that was set
        """
        if key is None:
            self.data.append(value)
        else:
            self.data[key] = value
        return value
    
    @classmethod
    def load_records(cls, records: typing.List[typing.Dict]) -> 'DataList[S]':
        """Load the struct list from records

        Args:
            records (typing.List[typing.Dict]): The list of records to load

        Returns:
            StructList[S]: The list of structs
        """
        structs = []
        struct_cls: typing.Type[pydantic.BaseModel] = generic_class(S)
        for record in records:
            structs.append(struct_cls.load(record))
        return DataList[S](
            structs=structs
        )


class MediaMessage(Message):
    """A message that contains media such as an image
    """

    def __init__(self, source: str, media: typing.List[Media]):
        """Create a media message with a source and a media

        Args:
            source (str): The source for the message
            media (typing.List[Media]): The media to use
        """
        super().__init__(
            source=source,
            data={
                'media': media
            }
        )

    def render(self) -> str:
        """Render the media message

        Returns:
            str: The rendered message
        """
        return f'{self.source}: Media [{self.media}]'



from typing import AsyncIterator, Dict
import pkg_resources
import openai
import typing

from .._core import Message, AIPrompt
from .._core import AIModel, AIResponse, Cue, TextMessage

# TODO: add utility for this
required = {'openai'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed


if len(missing) > 0:

    raise RuntimeError(f'To use this module openai must be installed.')


class OpenAIChatModel(AIModel):
    """A model that uses OpenAI's Chat API
    """

    def __init__(self, model: str, **kwargs) -> None:
        """Create an OpenAIChat model

        Args:
            model (str): The name of the model
        """
        super().__init__()
        self.model = model
        self.kwargs = kwargs

    def convert(self, message: Message) -> typing.Dict:
        """Convert the messages to a format to use for the OpenAI API

        Args:
            messages (typing.List[Message]): The messages

        Returns:
            typing.List[typing.Dict]: The input to the API
        """
        text = message['text']
        if isinstance(text, Cue):
            text = text.render()
        return {
            'role': message.source,
            'content': text
        }

    def forward(self, prompt: AIPrompt, **kwarg_override) -> AIResponse:
        """Execute the model

        Args:
            prompt (AIPrompt): The message to send the model

        Returns:
            AIResponse: the response from the model
        """
        kwargs = {
            **self.kwargs,
            **kwarg_override
        }
        client = openai.OpenAI()

        response = client.chat.completions.create(
            model=self.model,
            messages=self.convert_messages(prompt.aslist()),
            **kwargs
        )
        p = prompt.reader()
        text = response.choices[0].message.content
        message = TextMessage('assistant', text)
        
        val = p.read(text)
        
        return AIResponse(
            message,
            response,
            val
        )

    def stream_forward(
        self, prompt: AIPrompt, 
        **kwarg_override
    ) -> typing.Iterator[typing.Tuple[AIResponse, AIResponse]]:
        """Stream the model

        Args:
            prompt (AIPrompt): the model prmopt

        Yields:
            Iterator[typing.Iterator[typing.Tuple[AIResponse, AIResponse]]]: the responses from the model
        """
        kwargs = {
            **self.kwargs,
            **kwarg_override
        }
        
        client = openai.OpenAI()
        query = client.chat.completions.create(
            model=self.model,
            messages=self.convert_messages(prompt.aslist()),
            stream=True,
            **kwargs
        )
        cur_message = ''
        p = prompt.reader()
        for chunk in query:
            delta = chunk.choices[0].delta.content

            if delta is None:
                delta = ''
            
            cur_message = cur_message + delta
            
            message = TextMessage('assistant', cur_message)
            dx = TextMessage('assistant', delta)

            dx_val = p.read(delta)
            yield AIResponse(
                message, chunk, p.read(cur_message)
            ), AIResponse(
                dx, chunk, dx_val
            )
    
    async def async_forward(
        self, prompt: AIPrompt, 
        **kwarg_override
    ) -> typing.Tuple[str, typing.Dict]:
        """

        Args:
            prompt (AIPrompt): The 

        Returns:
            typing.Tuple[str, typing.Dict]: _description_
        """
        client = openai.AsyncOpenAI()

        kwargs = {
            **self.kwargs,
            **kwarg_override
        }
        p = prompt.reader()

        response = await client.chat.completions.create(
            model=self.model,
            messages=self.convert(prompt),
            **kwargs
        )
        text = response.choices[0].message.content
        return AIResponse(
            TextMessage('assistant', text),
            response,
            p.read(text)
        )

    async def async_stream_forward(self, prompt: AIPrompt, **kwarg_override) -> AsyncIterator[typing.Tuple[AIResponse, AIResponse]]:
        """Stream the model asyncrhonously

        Args:
            prompt (AIPrompt): The prompt for the model

        Yields:
            Iterator[AsyncIterator[typing.Tuple[AIResponse, AIResponse]]]: The response from the model
        """

        kwargs = {
            **self.kwargs,
            **kwarg_override
        }
        client = openai.AsyncOpenAI()
        query = await client.chat.completions.create(
            model=self.model,
            messages=self.convert(prompt.aslist()),
            stream=True,
            **kwargs
        )
        p = prompt.reader()

        cur_message = ''
        async for chunk in query:
            delta = chunk.choices[0].delta.content
            if delta is None:
                delta = ''
            cur_message = cur_message + delta

            yield AIResponse(
                TextMessage('assistant', cur_message),
                chunk,
                p.stream_read(cur_message),
            ), AIResponse(
                TextMessage('assistant', delta),
                chunk,
                None
            )


class OpenAIMessage(TextMessage):
    """A text message made to work with OpenAI. Especially makes
    using prompt more straightforward
    """
    
    def prompt(self, model: typing.Union[AIModel, str], **kwarg_overrides) -> AIResponse:
        """prompt the model

        Args:
            model (typing.Union[AIModel, str]): The model to execute

        Returns:
            AIResponse: The response from the model
        """
        if isinstance(model, str):
            model = OpenAIChatModel(model, **kwarg_overrides)

        return super().prompt(model, **kwarg_overrides)


class OpenEmbeddingModel(AIModel):

    def __init__(self, model: str, **kwargs) -> None:
        super().__init__()
        self.model = model
        self.kwargs = kwargs

    def convert(self, message: Message) -> typing.Dict:
        """Convert the messages to a format to use for the OpenAI API

        Args:
            messages (typing.List[Message]): The messages

        Returns:
            typing.List[typing.Dict]: The input to the API
        """
        return list(
            render(text_i) for text_i in message['texts']
        )


    def forward(self, prompt: AIPrompt, **kwarg_override) -> Dict:

        kwargs = {
            **self.kwargs,
            **kwarg_override
        }
        client = openai.OpenAI()

        response = client.embeddings.create(
            model=self.model,
            input=self.convert(prompt),
            **kwargs
        )
        embeddings = [e for e in response.data]
        
        return AIResponse(
            embeddings,
            response,
            embeddings
        )

    def stream_forward(
        self, prompt: AIPrompt, 
        **kwarg_override
    ) -> typing.Iterator[typing.Tuple[AIResponse, AIResponse]]:

        response = self.forward(prompt, **kwarg_override)
        yield response, response
        
    async def async_forward(
        self, prompt: AIPrompt, 
        **kwarg_override
    ) -> typing.Tuple[str, typing.Dict]:
        client = openai.AsyncOpenAI()

        kwargs = {
            **self.kwargs,
            **kwarg_override
        }
        response = client.embeddings.create(
            model=self.model,
            input=self.convert(prompt),
            **kwargs
        )
        embeddings = [e for e in response.data]
        
        return AIResponse(
            embeddings,
            response,
            embeddings
        )

    async def async_stream_forward(self, prompt: AIPrompt, **kwarg_override) -> AsyncIterator[typing.Tuple[AIResponse, AIResponse]]:

        response = await self.async_forward(prompt, **kwarg_override)
        yield response, response



from ._lang import (
    Term, Glossary
)
from ._data import (
    Context, ContextStorage,
    Shared, get_or_set, get_or_spawn, SharedBase,
    Buffer, BufferIter, ContextSpawner
)
from ._structs import (
    Media, 
    Message, 
    DataList, MediaMessage
)



from abc import ABC, abstractmethod
import json
from dataclasses import dataclass
import inspect
import typing
from typing import Self, get_type_hints
import pydantic
from ._utils import unescape_curly_braces, escape_curly_braces


class StructLoadException(Exception):
    """Exception StructLoad
    """

    def __init__(self, message="Struct loading failed.", errors=None):
        """Create a StructLoadException with a message

        Args:
            message (str, optional): The message. Defaults to "Struct loading failed.".
            errors (optional): The errors. Defaults to None.
        """
        super().__init__(message)
        self.errors = errors



class Renderable(ABC):
    """Mixin for classes that implement the render()
    method. Render is used to determine how to represent an
    object as a string to send to thte LLM
    """

    @abstractmethod
    def render(self) -> str:
        """Convert an object to a string representation for 
        an llm

        Returns:
            str: the string representation of the object
        """
        pass


class Templatable(ABC):
    """A mixin to indicate that the class 
    has a template function defined. Templates are
    used by the LLM to determine how to output.
    """

    @abstractmethod
    def template(self) -> str:
        """Get the template 

        Returns:
            str: 
        """
        pass



@dataclass
class TemplateField(Renderable):
    """Use for rendering a field in a BaseModel
    """
    type_: str
    description: str
    default: typing.Any = None
    is_required: bool = True

    def to_dict(self) -> typing.Dict:
        """Convert the template to a dict

        Returns:
            typing.Dict: the template
        """
        return {
            'type': self.type_,
            'description': self.description,
            'default': self.default,
            'is_required': self.is_required
        }
    
    def render(self) -> str:
        """Convert the template to a string

        Returns:
            str: The string of the template.
        """
        return str(self.to_dict())


def model_template(model_cls: typing.Type[pydantic.BaseModel]) -> str:
    """Get the template for a pydantic.Model

    Args:
        model_cls (typing.Type[pydantic.BaseModel]): The model to retrieve for

    Returns:
        str: The model template string
    """
    template = {}
    for name, field_type in get_type_hints(model_cls).items():
        
        if inspect.isclass(field_type) and issubclass(field_type, pydantic.BaseModel):
            template[name] = model_template(field_type)
        else:
            template[name] = {
                "is_required": model_cls.model_fields[name].is_required(),
                "type": field_type
            }
    return template


def struct_template(model: pydantic.BaseModel) -> typing.Dict:
    """Get the template for the Struct

    Returns:
        typing.Dict: The template 
    """
    template = {}
    
    base_template = model_template(model)
    for field_name, field in model.model_fields.items():
        field_type = field.annotation
        if isinstance(field_type, type) and issubclass(field_type, pydantic.BaseModel):

            template[field_name] = struct_template(field_type)
        else:

            if 'is_required' in base_template[field_name]:
                is_required = base_template[field_name]['is_required']
            else:
                is_required = True
            template[field_name] = TemplateField(
                type_=field.annotation,
                description=field.description,
                default=field.default if field.default is not None else None,
                is_required=is_required
            )

    return template


def model_to_text(model: pydantic.BaseModel, escape: bool=False) -> str:
    """Dump the struct to a string

    Returns:
        str: The string
    """
    if escape:  
        return escape_curly_braces(model.model_dump())
    return model.model_dump_json()


def model_from_text(model_cls: typing.Type[pydantic.BaseModel], data: str, escaped: bool=False) -> Self:
    """Load the struct from a string

    Args:
        data (str): The data for the struct

    Returns:
        Self: The loaded struct
    """
    if escaped:
        data = unescape_curly_braces(data)
    return model_cls(**json.loads(data))


def doc(obj) -> str:
    """utility to get the docstring for the object

    Args:
        obj: the object to get the docstring for

    Returns:
        str: The docstring for the object
    """
    d = obj.__doc__
    return d if d is not None else ''



# 1st party
import typing
import asyncio
from functools import partial
import threading
import time

# local
from ._core import (
    TaskStatus, Task
)
from ..data import Context, ContextSpawner, SharedBase

from ._core import TOSTATUS

from ..data import Buffer,Shared
from .._core import AIModel, AIPrompt


TASK = typing.Union[Task, typing.Callable[[typing.Dict], TaskStatus]]
CALL_TASK = typing.Callable[[],TaskStatus]


PARALLEL = typing.Callable[[typing.Iterable[Task], int, int, bool], TaskStatus]


async def _parallel(
    tasks: typing.Iterable[TASK], 
    success_on: int=-1, 
    fails_on: int=1, success_priority: bool=True
) -> TaskStatus:
    """Run in parallel

    Args:
        tasks (typing.List[TASK]): The tasks to run
        success_on (int): The number required for success
        fails_on (int): The number 
        success_priority (bool): The 

    Raises:
        ValueError: If the value is not defined

    Returns:
        TaskStatus: The resulting status
    """
    tg_tasks = []
    async with asyncio.TaskGroup() as tg:

        for _, task in enumerate(tasks):

            tg_tasks.append(tg.create_task(
                (asyncio.to_thread(task))
            ))

        if success_on < 0:
            success_on = len(tg_tasks) + 1 + success_on

        if fails_on < 0:
            fails_on = len(tg_tasks) + 1 + fails_on

        if (fails_on + success_on) > (len(tg_tasks) + 1):
            raise ValueError(
                f'Success + failure must be lte the number '
                f'of tasks not {fails_on + success_on}'
            )

    failed = 0
    succeeded = 0

    for tg_task in tg_tasks:
        cur_status = tg_task.result()
        if cur_status.success:
            succeeded += 1
        if cur_status.failure:
            failed += 1

    if success_priority:
        if succeeded >= success_on:
            return TaskStatus.SUCCESS
        if failed >= fails_on:
            return TaskStatus.FAILURE
    else:
        if failed >= fails_on:
            return TaskStatus.FAILURE
        if succeeded >= success_on:
            return TaskStatus.SUCCESS
    return TaskStatus.RUNNING


def parallel(
    tasks: typing.Iterable[TASK], 
    succeeds_on: int=-1, 
    fails_on: int=1, 
    success_priority: bool=True,
    parallelizer: PARALLEL=None
) -> CALL_TASK:
    """Create a parallel task

    Args:
        tasks (typing.Iterable[TASK]): The tasks to run
        succeeds_on (int, optional): The number of Successes required to succeed. Defaults to -1.
        fails_on (int, optional): The number of Failures required to fail. Defaults to 1.
        success_priority (bool, optional): Whether success is prioritized over failure. Defaults to True.

    Returns:
        CALL_TASK: The task to call
    """
    def _f():
        if parallelizer:
            return parallelizer(tasks, succeeds_on, fails_on, success_priority)
        
        return asyncio.run(
            _parallel(tasks, succeeds_on, fails_on, success_priority)
        )
    return _f


def parallelf(
    f: typing.Callable[[typing.Any], typing.Iterator[TASK]],
    *args, 
    succeeds_on: int=-1, 
    fails_on: int=1, 
    success_priority: bool=True,
    **kwargs
) -> CALL_TASK:
    """Create a parallel task

    Args:
        tasks (typing.Iterable[TASK]): The tasks to run
        succeeds_on (int, optional): The number of Successes required to succeed. Defaults to -1.
        fails_on (int, optional): The number of Failures required to fail. Defaults to 1.
        success_priority (bool, optional): Whether success is prioritized over failure. Defaults to True.

    Returns:
        CALL_TASK: The task to call
    """
    return parallel(
        f(*args, **kwargs), succeeds_on, fails_on, success_priority
    )


def spawn(
    f, n: int,
    *args, 
    **kwargs
) -> typing.List[TASK]:
    """Convert a task into multiple tasks by spawning multiple
    states. Any "states" that are included in args or kwargs

    Args:
        f: The function to execute
        n (int): The number of tasks to spawn

    Returns:
        typing.List[TASK]: The list of tasks to run
    """

    tasks = []

    for i in range(n):
        cur_args = [
            arg[i] if isinstance(arg, ContextSpawner) else arg
            for arg in args.items()
        ]
        cur_kwargs = {
            k: arg[i] if isinstance(arg, ContextSpawner) else arg
            for k, arg in kwargs.items()
        }

        tasks.append(partial(f, *cur_args, **cur_kwargs))

    return tasks


def sequence(tasks: typing.Iterable[TASK], ctx: Context) -> CALL_TASK:
    """Run a sequence task

    Args:
        tasks (typing.Iterable[TASK]): The tasks to execute
        state (State): The current state

    Returns:
        CALL_TASK: The task to call
    """
    def _f():
        status = ctx.get_or_set('status', TaskStatus.RUNNING)

        if status.is_done:
            return status
        
        # get the iterator
        if 'it' not in ctx:
            ctx['it'] = iter(tasks)

            try:
                cur_task = next(ctx['it'])
                ctx['cur_task'] = cur_task
            except StopIteration:
                ctx['status'] = TaskStatus.SUCCESS
                return TaskStatus.SUCCESS
            
        cur_task = ctx['cur_task']

        if isinstance(cur_task, bool):
            cur_status = TaskStatus.from_bool(cur_task)
        elif cur_task == TaskStatus.FAILURE or cur_task == TaskStatus.SUCCESS:
            cur_status = cur_status
        # It must be a task
        else:
            cur_status = cur_task()

        # update the sequence status
        if cur_status.running:
            ctx['status'] = TaskStatus.RUNNING
            return ctx['status']

        if cur_status.success:

            try:
                cur_task = next(ctx['it'])
                ctx['cur_task'] = cur_task
                ctx['status'] = TaskStatus.RUNNING
            except StopIteration:
                ctx['status'] = TaskStatus.SUCCESS
                return TaskStatus.SUCCESS
        else:
            ctx['status'] = cur_status

        return ctx['status']

    return _f


def sequencef(f: typing.Callable[[typing.Any], typing.Iterator[TASK]], ctx: Context, *args, **kwargs) -> CALL_TASK:
    """Run a callable sequence task

    Args:
        tasks (typing.Iterable[TASK]): The tasks to execute
        state (State): The current state

    Returns:
        CALL_TASK: The task to call
    """
    return sequence(f(*args, **kwargs), ctx)


# SELECTOR does not seem to be working correctly

def _selector(
    tasks: typing.List[TASK], 
    ctx: Context
) -> TaskStatus:

    status = ctx.get_or_set('status', TaskStatus.RUNNING)

    if status.is_done:
        return status
    
    # get the iterator
    if 'it' not in ctx:
        ctx['it'] = iter(tasks)

        try:
            cur_task = next(ctx['it'])
            ctx['cur_task'] = cur_task
        except StopIteration:
            ctx['status'] = TaskStatus.FAILURE
            return TaskStatus.FAILURE
        
    cur_task = ctx['cur_task']
    
    if isinstance(cur_task, bool):
        cur_status = TaskStatus.from_bool(cur_task)
    elif cur_task == TaskStatus.FAILURE or cur_task == TaskStatus.SUCCESS:
        cur_status = cur_status
    # It must be a task
    else:
        cur_status = cur_task()

    if cur_status.running:
        ctx['status'] = TaskStatus.RUNNING
        return ctx['status']

    if cur_status.failure:

        try:
            cur_task = next(ctx['it'])
            ctx['cur_task'] = cur_task
            ctx['status'] = TaskStatus.RUNNING
        except StopIteration:
            ctx['status'] = TaskStatus.FAILURE
            return TaskStatus.FAILURE
    else:
        ctx['status'] = cur_status

    return ctx['status']


def selector(tasks: TASK, ctx: Context) -> CALL_TASK:
    """Create a selector task

    Args:
        tasks (TASK): The tasks to run
        state (State): The state for the selector

    Returns:
        CALL_TASK: The task to call
    """
    def _f():
        return _selector(tasks, ctx)

    return _f


def selectorf(
    f: typing.Callable[[typing.Any], typing.Iterator[TASK]], 
    ctx: Context, *args, **kwargs) -> CALL_TASK:
    """Run a callable selector task

    Args:
        tasks (typing.Iterable[TASK]): The tasks to execute
        state (State): The current state

    Returns:
        CALL_TASK: The task to call
    """
    return selector(f(*args, **kwargs), ctx)


fallback = selector
fallbackf = selectorf


def action(task: TASK, *args, **kwargs) -> CALL_TASK:
    """Functional form of action

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute
        state (typing.Dict): The state of execution

    Returns:
        TaskStatus: The status of the result
    """
    def _f():
        result = task(*args, **kwargs)
        return result

    return _f


def taskf(f, *args, out: SharedBase=None, to_status: TOSTATUS=None, **kwargs) -> CALL_TASK:
    """A generic task based on a function

    Args:
        f: The function to exectue
        out (SharedBase, optional): The output to store to. Defaults to None.
        to_status (TOSTATUS, optional): The status converter. If not set will automatically set to "Success". 
            Defaults to None.

    Returns:
        CALL_TASK: The CALL_TASK to execute
    """
    def _f():
        result = f(*args, **kwargs)
        if out is not None:
            out.set(result)
        
        if to_status is not None:
            status = to_status(result)
        elif isinstance(result, TaskStatus):
            status = result
        else:
            status = TaskStatus.SUCCESS
        return status

    return _f


def condf(task: typing.Callable, *args, **kwargs) -> CALL_TASK:
    """Functional form of action

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute
        state (typing.Dict): The state of execution

    Returns:
        TaskStatus: The status of the result
    """
    def _f():
        result = task(*args, **kwargs)
        return TaskStatus.from_bool(result)

    return _f


def not_(task: TASK) -> CALL_TASK:
    """Invert the result of the task if Failure or Success

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute
        state (typing.Dict): The state of execution

    Returns:
        TaskStatus: The status of the result
    """
    def _f() -> TaskStatus:
        status = tick(task)
        return status.invert()

    return _f


def notf(f, *args, **kwargs) -> CALL_TASK:
    """Invert the result of the task if Failure or Success

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute
        state (typing.Dict): The state of execution

    Returns:
        TaskStatus: The status of the result
    """
    return not_(partial(f, *args, **kwargs))


def _run_thread(task: TASK, ctx: Context, interval: float=1./60):
    """Run periodically to update the status

    Args:
        task (TASK): The task to run
        ctx (Context): The context
        interval (float, optional): The interval to run at. Defaults to 1./60.
    """
    while True:
        status = task()
        ctx['thread_status'] = status
        if status.is_done:
            break
        time.sleep(interval)


def threaded(task: TASK, ctx: Context, interval: float=1./60) -> CALL_TASK:

    def run() -> TaskStatus:
        if '_thread' not in ctx:
            ctx['thread_status'] = TaskStatus.RUNNING
            t = threading.Thread(target=_run_thread, args=(task, ctx, interval))
            t.start()
            ctx['_thread'] = t
        
        return ctx['thread_status']

    if 'task_id' in ctx and id(task) != ctx['task_id']:

        raise RuntimeError(
            'Task context has been initialized but '
            'the task passed in is does not match'
        )

    return run


def _stream_model(model: AIModel, prompt: AIPrompt, ctx: Context, *args, interval: float=1./60, **kwargs):
    """Run periodically to update the status

    Args:
        task (TASK): The task to run
        ctx (Context): The context
        interval (float, optional): The interval to run at. Defaults to 1./60.
    """
    print('Executing thread')
    for x, dx in model.stream_forward(prompt, *args, **kwargs):
        ctx['x'] = x
        ctx['dx'].append(dx)
        time.sleep(interval)
    ctx['thread_status'] = TaskStatus.SUCCESS


def stream_model(
    buffer: Buffer, engine: AIModel, prompt: AIPrompt, ctx: Context, 
    *args, interval: float=1./60,  **kwargs
) -> CALL_TASK:
    """Execute the AI model in a thread

    Args:
        shared (Shared): THe shared
        engine (AIModel): The model to use
        ctx (Context): The context to use for maintaining state

    Returns:
        CALL_TASK
    """
    def run() -> TaskStatus:
        if '_thread' not in ctx:
            ctx['x'] = None
            ctx['dx'] = []
            ctx['i'] = 0
            ctx['thread_status'] = TaskStatus.RUNNING
            print('create thread')
            t = threading.Thread(
                target=_stream_model, args=(
                    engine, prompt, ctx, *args
                ), kwargs={'interval': interval, **kwargs}
            )
            t.start()
            ctx['_thread'] = t
        
        if ctx['i'] < len(ctx['dx']):
            buffer.add(*ctx['dx'][ctx['i']:])
            ctx['i'] = len(ctx['dx'])

        return ctx['thread_status']

    return run


def _run_model(model: AIModel, prompt: AIPrompt, ctx: Context, **kwargs):
    """Run periodically to update the status

    Args:
        task (TASK): The task to run
        ctx (Context): The context
        interval (float, optional): The interval to run at. Defaults to 1./60.
    """
    ctx['x'] = model(prompt, **kwargs)
    ctx['thread_status'] = TaskStatus.SUCCESS


def exec_model(
    shared: Shared, engine: AIModel, prompt: AIPrompt, ctx: Context, 
    **kwargs
) -> CALL_TASK:
    """Execute the AI model in a thread

    Args:
        shared (Shared): THe shared
        engine (AIModel): The model to use
        ctx (Context): The context to use for maintaining state

    Returns:
        CALL_TASK
    """
    def run() -> TaskStatus:
        if '_thread' not in ctx:
            ctx['x'] = None
            ctx['thread_status'] = TaskStatus.RUNNING
            t = threading.Thread(
                target=_run_model, args=(
                    engine, prompt, ctx
                ), kwargs=kwargs
            )
            t.start()
            ctx['_thread'] = t
        
        if ctx['thread_status'].is_done:
            shared.set(ctx['x'])

        return ctx['thread_status']

    return run


def _run_func(ctx: Context, f, *args, **kwargs):
    """Run periodically to update the status

    Args:
        task (TASK): The task to run
        ctx (Context): The context
        interval (float, optional): The interval to run at. Defaults to 1./60.
    """
    ctx['x'] = f(*args, **kwargs)
    ctx['thread_status'] = TaskStatus.SUCCESS


def exec_func(
    shared: Shared, ctx: Context, f: typing.Callable,
    *args,
    **kwargs
) -> CALL_TASK:
    """Execute the AI model in a thread

    Args:
        shared (Shared): THe shared
        engine (AIModel): The model to use
        ctx (Context): The context to use for maintaining state

    Returns:
        CALL_TASK
    """
    def run() -> TaskStatus:
        if '_thread' not in ctx:
            ctx['x'] = None
            ctx['thread_status'] = TaskStatus.RUNNING
            t = threading.Thread(
                target=_run_func, args=(
                    ctx, f, *args
                ), kwargs=kwargs
            )
            t.start()
            ctx['_thread'] = t
        
        if ctx['thread_status'].is_done:
            shared.set(ctx['x'])

        return ctx['thread_status']

    return run


def tick(task: TASK) -> TaskStatus:
    """Run the task

    Args:
        task (TASK): The task to run

    Returns:
        TaskStatus: The resulting status from the task
    """

    if isinstance(task, Task):
        return task.tick()
    
    return task()


def unless(task: TASK, status: TaskStatus=TaskStatus.FAILURE) -> CALL_TASK:
    """Loop unless a condition is met

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute

    Returns:
        TaskStatus: The status of the result
    """
    def _f():
        cur_status = task()
        
        if cur_status == status:
            return TaskStatus.RUNNING
        return cur_status
    return _f


def unlessf(f, *args, status: TaskStatus=TaskStatus.FAILURE, **kwargs) -> CALL_TASK:
    """Loop unless a condition is met

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute

    Returns:
        CALL_TASK: The status of the result
    """
    return unless(partial(f, *args, **kwargs), status)


def until(task: TASK, status: TaskStatus=TaskStatus.SUCCESS) -> CALL_TASK:
    """Loop until a condition is met

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute

    Returns:
        TaskStatus: The status of the result
    """

    def _f():
        cur_status = task()
        
        if cur_status == status:
            return cur_status
        return TaskStatus.RUNNING
    return _f


def untilf(f, *args, status: TaskStatus=TaskStatus.SUCCESS, **kwargs) -> CALL_TASK:
    """Loop until a condition is met

    Args:
        task (typing.Union[_tasks.Task, typing.Callable[[], TaskStatus]]): the task to execute

    Returns:
        TaskStatus: The status of the result
    """
    return until(partial(f, *args, **kwargs), status)



# 1st party
import typing
import string

import pydantic
import roman

# local
from .._core import (
    render, render_multi,
    Cue, render, Param, validate_out 
)
from ..utils import (
    str_formatter
)
from .._core import Module
from ._data import Description


S = typing.TypeVar('S', bound=pydantic.BaseModel)
X = typing.Union[str, Description, Cue]


def bullet(xs: typing.Iterable[X], bullets: str='-', indent: int=0) -> 'Cue':
    """Create a bullet list based on the instructions

    Args:
        xs (typing.Iterable[X]): The cues to bullet
        bullets (str, optional): The string to use for the bullet. Defaults to '-'.

    Returns:
        Cue: The resulting cue
    """
    indent = ' ' * indent
    text = f'\n{indent}{bullets}'
    out = validate_out(xs)
    text = text + f'\n{indent}{bullets}'.join(
        render(x_i) for x_i in xs
    )
    return Cue(
        text=text, out=out
    )


def formatted(x: X, format: str) -> 'Cue':
    """Format the X with a format. The format string will encapsulate it

    Example:

    formatted('Name', '*') => '*Name*'

    Args:
        x (X): The data to format
        format (str): The format to use

    Returns:
        Cue: The resulting cue
    """

    text = render(x)
    if text[:len(format)] == format and text[-len(format):] == format:
        return x
    return Cue(
        f'{format}{text}{format}',
        out=x.out
    )


def bold(x: X) -> 'Cue':
    """Format the X with a bold format. The format string will encapsulate it

    Example:

    bold('Name') => '**Name**'

    Args:
        x (X): The data to format

    Returns:
        Cue: The resulting cue
    """

    return formatted(x, '**')


def italic(x: X) -> 'Cue':
    """Format the X with a bold format. The format string will encapsulate it

    Example:

    italic('Name') => '*Name*'

    Args:
        x (X): The data to format

    Returns:
        Cue: The resulting cue
    """
    return formatted(x, '*')


def strike(x: X) -> 'Cue':
    """Format the X with a bold format. The format string will encapsulate it

    Example:

    italic('Name') => '*Name*'

    Args:
        x (X): The data to format

    Returns:
        Cue: The resulting cue
    """
    return formatted(x, '~~')


def generate_numbered_list(n, numbering_type='arabic') -> typing.List:
    """Generate a numbered list in arabic or roman numerals

    Args:
        n: The number of numbers to output
        numbering_type (str, optional): The type of numberals to use
        "arabic" or "roman" or "alphabet". Defaults to 'arabic'.

    Raises:
        ValueError: If the amount of numbers is not supported
        ValueError: If the numbering system is incorrect

    Returns:
        _type_: _description_
    """
    if numbering_type == 'arabic':
        return [str(i) for i in range(1, n + 1)]
    elif numbering_type == 'roman':
        return [roman.toRoman(i).lower() for i in range(1, n + 1)]
    elif numbering_type == 'alphabet':
        if n > 26:
            raise ValueError("Alphabetic numbering can only handle up to 26 items")
        return [string.ascii_uppercase[i] for i in range(n)]
    else:
        raise ValueError("Unsupported numbering type")


def numbered(xs: typing.Iterable[X], indent: int=0, numbering: str='arabic') -> 'Cue':
    """Create a numbered list

    Args:
        xs (typing.Iterable[Cue]): A list of strings
        indent (int, optional): The number to start with indenting. Defaults to 0.
        numbering (str, optional): The type of numbering system to use. Defaults to 'arabic'.

    Returns:
        Cue: The resulting Cue
    """
    text = ''
    indent = ' ' * indent
    numbers = generate_numbered_list(len(xs), numbering)
    out = validate_out(xs)
    for i, (x_i, number) in enumerate(zip(xs, numbers)):
        text = f'{indent}{number}. {render(x_i)}'
        if i < (len(numbers) - 1):
            text += "\n"

    return Cue(
        text=text, 
        out=out
    )


def fill(x_instr: X, *args: X, **kwargs: X) -> 'Cue':
    """Format a string with variables

    Args:
        x_instr (X): The value to format

    Returns:
        Cue: The resulting cue
    """
    out = validate_out([x_instr])

    kwargs = dict(zip(kwargs.keys(), render_multi(kwargs.values())))
    args = render_multi(args)
    return Cue(
        text=str_formatter(render(x_instr), *args, **kwargs), out=out
    )


def head(x: X, size: int=1) -> 'Cue':
    """Add a header to the cue

    Args:
        x (X): The input to add to
        size (int, optional): The size of the heading. Defaults to 1.

    Returns:
        Cue: The resulting cue
    """
    out = validate_out([x])
    heading = '#' * size
    return Cue(
        text=f'{heading} {render(x)}', out=out
    )


def section(name: X, details: X, size: int=1, linebreak: int=1) -> 'Cue':
    """Add a section to the cue

    Args:
        name (X): The name of the section
        details (X): The details for the section
        size (int, optional): The size of the heading. Defaults to 1.
        linebreak (int, optional): How many linebreaks to put between the heading and the details. Defaults to 1.

    Returns:
        Cue: The inst
    """
    heading = '#' * size
    out = validate_out([name, details])
    linebreak = '\n' * linebreak
    text = f'{heading} {render(name)}{linebreak}' + render(details)
    return Cue(
        text=text, out=out
    )


def cat(xs: typing.List[Cue], sep: str=' ') -> Cue:
    """Concatenate multiple cues together

    Args:
        xs (typing.List[Cue]): THe cues 
        sep (str): The delimiter to use for the sections

    Raises:
        RuntimeError: 

    Returns:
        Cue: The concatenated cues
    """
    out = validate_out(xs)

    return Cue(text=f'{sep}'.join(
        render(x_i) for x_i in xs
    ), out=out)


def join(x1: X, x2: X, sep: str=' ') -> Cue:
    """Join two instructions together

    Args:
        x1 : Cue 1
        x2 : Cue 2
        sep (str): The separator between the two instructions

    Returns:
        Cue: The joined instructions
    """
    out = validate_out([x1, x2])
    return Cue(
        text=render(x1) + sep + render(x2),
        out=out
    )


class Op(Module):
    """An operation acts on an cue to produce a new cue
    """

    def __init__(
        self, name: str, cue: X, 
        tunable: bool=False):
        """Create an operation specifying the name

        Args:
            name (str): The name of the operation
            cue (X): The cue for the operation
            tunable: whether the cue is tunable
        """
        self.name = name
        if not isinstance(cue, Cue):
            cue = Param(
                name=self.name, training=tunable, 
                cue=Cue(
                    text=render(cue)
            ))
        
        self.cue = cue
        
    def forward(
        self, *args: X, **kwargs: X
    ) -> Cue:
        """Fill in the cue with the inputs

        Returns:
            Cue: 
        """
        cue = render(self.cue)
        out = validate_out(
            [*args, *kwargs.values(), self.cue]
        )

        return Cue(
            text=fill(cue, *args, **kwargs), out=out
        )


def op(x: typing.Union[typing.Iterable[X], X], cue: X) -> Cue:
    """Execute an operation on a cue

    Args:
        x (typing.Union[typing.Iterable[X], X]): The input
        cue (X): The cue for the operation

    Returns:
        Cue: The resulting cue
    """
    if not isinstance(x, typing.Iterable):
        x = [x]

    out = validate_out([*x, cue])
    resources = ', '.join(render_multi(x))
    # resources = ', '.join(x_i.name for x_i in x)
    text = f'Do: {render(cue)} --- With Inputs: {resources}'
    return Cue(
        text=text, out=out
    )



from abc import abstractmethod, ABC
import pydantic
from ..utils import Renderable


class Description(pydantic.BaseModel, Renderable, ABC):
    """Provide context in the prompt template
    """
    name: str = pydantic.Field(description='The name of the description.')

    @abstractmethod
    def render(self) -> str:
        pass


class Ref(pydantic.BaseModel, Renderable):
    """Reference to another description.
    Useful when one only wants to include the 
    name of a description in part of the prompt
    """
    desc: Description

    @property
    def name(self) -> str:
        """Get the name of the ref

        Returns:
            str: The name of the ref
        """
        return self.desc.name

    def render(self) -> str:
        """Generate the text rendering of the ref

        Returns:
            str: The name for the ref
        """
        return self.desc.name



# 1st party
from enum import Enum
from abc import abstractmethod
import typing
from dataclasses import dataclass

# local
from .._core import Storable


class TaskStatus(Enum):
    """Status of a Behavior Tree task
    """
    # The task is currently in progress
    RUNNING = 'running'
    # The task is currently waiting for a response
    WAITING = 'waiting'
    # The task was successful
    SUCCESS = 'success'
    # The task failed
    FAILURE = 'failure'
    # The task has not been executed
    READY = 'ready'

    @property
    def is_done(self) -> bool:
        """Get whether the task has succeeded or failed

        Returns:
            bool: Whether the task is "done"
        """
        return self == TaskStatus.FAILURE or self == TaskStatus.SUCCESS
    
    @property
    def in_progress(self) -> bool:
        """Get whether the task is in progress

        Returns:
            bool: whether the task is still running
        """
        return self == TaskStatus.RUNNING or self == TaskStatus.WAITING
    
    @property
    def ready(self) -> bool:
        """Get whether the task is ready to be executed

        Returns:
            bool: Whether the task is ready to be executed
        """
        return self == TaskStatus.READY
    
    @property
    def failure(self) -> bool:
        """Get whether the task has failed

        Returns:
            bool: whether the task has failed
        """
        return self == TaskStatus.FAILURE
    
    @property
    def success(self) -> bool:
        """Get whether the task 

        Returns:
            bool: Whether the task has succeeded
        """
        return self == TaskStatus.SUCCESS
    
    @property
    def running(self) -> bool:
        """Get whether the task is running

        Returns:
            bool: Whether the task is running or not
        """
        return self == TaskStatus.RUNNING
    
    @classmethod
    def from_bool(cls, val: bool) -> 'TaskStatus':
        """Convert a boolean to a TaskStatus

        Args:
            val (bool): The value to convert

        Returns:
            TaskStatus: The status
        """
        return TaskStatus.SUCCESS if val is True else TaskStatus.FAILURE
    
    def __or__(self, other: 'TaskStatus') -> 'TaskStatus':
        """Compute the union of two statuses

        Returns:
            SangoStatus: The resulting status. Returns success if one status
            is success.
        """
        if self == other:
            return self
        
        if (
            (self.success or other.success)
        ):
            return TaskStatus.SUCCESS
        if self.running or other.running:
            return TaskStatus.RUNNING
        
        if self.waiting or other.waiting:
            return TaskStatus.WAITING
        
        if (self.success and not other.success):
            return other
        if (not self.failure and other.failure):
            return self
        
        raise ValueError(f'Invalid combination of statuses {self} and {other}')

    def __and__(self, other: 'TaskStatus') -> 'TaskStatus':
        """Compute the union of two statuses

        Returns:
            SangoStatus: The resulting status. Returns failure if one status
            is failure.
        """

        if self == other:
            return self
        
        if (
            (self.failure or other.failure)
        ):
            return TaskStatus.FAILURE
        if self.running or other.running:
            return TaskStatus.RUNNING
        
        if self.waiting or other.waiting:
            return TaskStatus.WAITING
        
        if (self.success and not other.success):
            return other
        if (not self.success and other.success):
            return self
        raise ValueError(f'Invalid combination of statuses {self} and {other}')

    def invert(self) -> 'TaskStatus':

        if self.success:
            return TaskStatus.FAILURE
        if self.failure:
            return TaskStatus.SUCCESS
        return self


WAITING = TaskStatus.WAITING
READY = TaskStatus.READY
SUCCESS = TaskStatus.SUCCESS
FAILURE = TaskStatus.FAILURE
RUNNING = TaskStatus.RUNNING

@dataclass
class TaskMessage:

    name: str
    data: typing.Any


class Task(Storable):
    """The base class for a task in the behavior tree
    """

    SUCCESS = TaskStatus.SUCCESS
    FAILURE = TaskStatus.FAILURE
    RUNNING = TaskStatus.RUNNING

    def __init__(self) -> None:
        """Create the task

        Args:
            name (str): The name of the task
        """
        super().__init__()
        self._status = TaskStatus.READY

    @abstractmethod    
    def tick(self) -> TaskStatus:
        raise NotImplementedError

    def __call__(self) -> TaskStatus:
        """Execute the task

        Returns:
            TaskStatus: The status of the task after execution
        """
        return self.tick()

    def reset(self):
        """Reset the terminal

        """
        self._status = TaskStatus.READY
    
    @property
    def status(self) -> TaskStatus:
        """The status

        Returns:
            TaskStatus: 
        """
        return self._status

    @property
    def id(self):
        """Get the id for the task

        Returns:
            The id of the task 
        """
        return self._id


class ToStatus(object):
    """Use to convert a value to a status
    """

    @abstractmethod
    def __call__(self, val) -> TaskStatus:
        """Convert the value to a status

        Args:
            val: The value to convert

        Returns:
            TaskStatus: The status
        """
        pass


TOSTATUS = ToStatus | typing.Callable[[typing.Any], TaskStatus]


def from_bool(status: bool) -> TaskStatus:
    """functional code to map from a boolean

    Args:
        status (bool): The status in boolean form

    Returns:
        TaskStatus: The task status
    """
    return TaskStatus.from_bool(status)


class State(object):
    """Use State creating a state machine
    """

    @abstractmethod
    def update(self) -> typing.Union['State', TaskStatus]:
        """Update the 

        Returns: bool
            typing.Union['State', TaskStatus]: The new State/Status
        """
        pass


class Router(object):
    """Use to a state
    """

    @abstractmethod
    def __call__(self, val) -> TaskStatus | State:
        pass


ROUTE = Router | typing.Callable[[typing.Any], TaskStatus | State]

# TODO: Add in Action (For GOAP)



from ._data import (
    Description, Ref   
)
from ._instruct import (
    bullet, formatted, generate_numbered_list,
    numbered, validate_out, fill, head,
    section, cat, join, Op, op,
    bold, strike, italic
)
